# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010                                                        *
# *   Yorik van Havre <yorik@uncreated.net>, Ken Cline <cline@frii.com>     *
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *   (c) 2020 Carlo Pavan <carlopa@gmail.com>                              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides utility functions that deal with GUI interactions.

This module contains auxiliary functions which can be used
in other modules of the workbench, and which require
the graphical user interface (GUI), as they access the view providers
of the objects or the 3D view.
"""
## @package gui_utils
# \ingroup draftutils
# \brief Provides utility functions that deal with GUI interactions.

## \addtogroup draftutils
# @{
import importlib
import math
import os

import FreeCAD as App
from draftutils import params
from draftutils import utils
from draftutils.messages import _err, _wrn
from draftutils.translate import translate

if App.GuiUp:
    import FreeCADGui as Gui
    from pivy import coin
    from PySide import QtCore
    from PySide import QtGui

    # from PySide import QtSvg  # for load_texture


def get_3d_view():
    """Return the current 3D view.

    Returns
    -------
    Gui::View3DInventor
        The Active 3D View or `None`.
    """
    if not App.GuiUp:
        return None

    # FIXME The following two imports were added as part of PR4926
    # Also see discussion https://forum.freecad.org/viewtopic.php?f=3&t=60251
    import FreeCADGui as Gui
    from pivy import coin

    mw = Gui.getMainWindow()
    view = mw.getActiveWindow()
    if view is None:
        return None
    if not hasattr(view, "getSceneGraph"):
        return None
    return view


get3DView = get_3d_view


def autogroup(obj):
    """Add a given object to the defined Draft autogroup, if applicable.

    This function only works if the graphical interface is available.
    It checks that the `App.draftToolBar` class is available,
    which contains the group to use to automatically store
    new created objects.

    Originally, it worked with standard groups (`App::DocumentObjectGroup`),
    and Arch Workbench containers like `'Site'`, `'Building'`, `'Floor'`,
    and `'BuildingPart'`.

    Now it works with Draft Layers.

    Parameters
    ----------
    obj: App::DocumentObject
        Any type of object that will be stored in the group.
    """

    # check for required conditions for autogroup to work
    if not App.GuiUp:
        return
    if not hasattr(Gui, "draftToolBar"):
        return
    if not hasattr(Gui.draftToolBar, "autogroup"):
        return
    if Gui.draftToolBar.isConstructionMode():
        return

    # check first for objects that do autogroup themselves
    # at the moment only Arch_BuildingPart, which is an App::GeometryPython
    for par in App.ActiveDocument.findObjects(Type="App::GeometryPython"):
        if hasattr(par.Proxy, "autogroup"):
            if par.Proxy.autogroup(par, obj):
                return

    # autogroup code
    if Gui.draftToolBar.autogroup is not None:
        active_group = App.ActiveDocument.getObject(Gui.draftToolBar.autogroup)
        if active_group is None:
            # Layer/group does not exist (anymore)
            Gui.draftToolBar.setAutoGroup()  # Change active layer/group in Tray to None.
        elif utils.get_type(active_group) == "Layer":
            if not obj in active_group.Group:
                active_group.Group += [obj]
            # No return statement here as objects can be in a layer and in
            # a normal group or group-like BIM object at the same time.
        elif obj in active_group.InListRecursive:
            return
        else:
            if not obj in active_group.Group:
                if hasattr(active_group, "addObject"):
                    active_group.addObject(obj)
                else:
                    active_group.Group += [obj]
            return

    if Gui.ActiveDocument.ActiveView.getActiveObject("NativeIFC") is not None:
        # NativeIFC handling
        try:
            from nativeifc import ifc_tools

            parent = Gui.ActiveDocument.ActiveView.getActiveObject("NativeIFC")
            ifc_tools.aggregate(obj, parent)
        except:
            pass

    elif Gui.ActiveDocument.ActiveView.getActiveObject("Arch") is not None:
        # add object to active Arch Container
        active_arch_obj = Gui.ActiveDocument.ActiveView.getActiveObject("Arch")
        if obj in active_arch_obj.InListRecursive:
            # do not autogroup if obj points to active_arch_obj to prevent cyclic references
            return
        active_arch_obj.addObject(obj)

    elif Gui.ActiveDocument.ActiveView.getActiveObject("part") is not None:
        # add object to active part and change it's placement accordingly
        # so object does not jump to different position, works with App::Link
        # if not scaled. Modified accordingly to realthunder suggestions
        active_part, parent, sub = Gui.ActiveDocument.ActiveView.getActiveObject("part", False)
        if obj in active_part.InListRecursive:
            # do not autogroup if obj points to active_part to prevent cyclic references
            return
        matrix = parent.getSubObject(sub, retType=4)
        if matrix.hasScale() == App.ScaleType.Uniform:
            err = translate("draft", "Unable to insert new object into " "a scaled part")
            App.Console.PrintMessage(err)
            return
        inverse_placement = App.Placement(matrix.inverse())
        if utils.get_type(obj) == "Point":
            point_vector = App.Vector(obj.X, obj.Y, obj.Z)
            real_point = inverse_placement.multVec(point_vector)
            obj.X = real_point.x
            obj.Y = real_point.y
            obj.Z = real_point.z
        elif utils.get_type(obj) in ["Dimension", "LinearDimension"]:
            obj.Start = inverse_placement.multVec(obj.Start)
            obj.End = inverse_placement.multVec(obj.End)
            obj.Dimline = inverse_placement.multVec(obj.Dimline)
            obj.Normal = inverse_placement.Rotation.multVec(obj.Normal)
            obj.Direction = inverse_placement.Rotation.multVec(obj.Direction)
        elif utils.get_type(obj) in ["Label"]:
            obj.Placement = App.Placement(inverse_placement.multiply(obj.Placement))
            obj.TargetPoint = inverse_placement.multVec(obj.TargetPoint)
        elif hasattr(obj, "Placement"):
            # every object that have a placement is processed here
            obj.Placement = App.Placement(inverse_placement.multiply(obj.Placement))

        active_part.addObject(obj)


def dim_symbol(symbol=None, invert=False):
    """Return the specified dimension symbol.

    Parameters
    ----------
    symbol: int, optional
        It defaults to `None`, in which it gets the value from the parameter
        database, `get_param("dimsymbolend")`.

        A numerical value defines different markers
         * 0, `SoSphere`
         * 1, `SoSeparator` with a `SoLineSet`, a circle (in fact a 24 sided polygon)
         * 2, `SoSeparator` with a `soCone`
         * 3, `SoSeparator` with a `SoFaceSet`
         * 4, `SoSeparator` with a `SoLineSet`, calling `dim_dash`
         * 5, Nothing
         * Otherwise, `SoSphere`

    invert: bool, optional
        It defaults to `False`.
        If it is `True` and `symbol=2`, the cone will be rotated
        -90 degrees around the Z axis, otherwise the rotation is positive,
        +90 degrees.

    Returns
    -------
    Coin.SoNode
        A `Coin.SoSphere`, or `Coin.SoSeparator` (circle, cone, face, line)
        that will be used as a dimension symbol.
    """
    if symbol is None:
        symbol = params.get_param("dimsymbolend")

    if symbol == 0:
        # marker = coin.SoMarkerSet()
        # marker.markerIndex = 80

        # Returning a sphere means that the bounding box will
        # be 3-dimensional; a marker will always be planar seen from any
        # orientation but it currently doesn't work correctly
        marker = coin.SoSphere()
        return marker
    elif symbol == 1:
        marker = coin.SoSeparator()
        v = coin.SoVertexProperty()
        for i in range(25):
            ang = math.radians(i * 15)
            v.vertex.set1Value(i, (math.sin(ang), math.cos(ang), 0))
        p = coin.SoLineSet()
        p.vertexProperty = v
        marker.addChild(p)
        return marker
    elif symbol == 2:
        marker = coin.SoSeparator()
        t = coin.SoTransform()
        t.translation.setValue((0, -2, 0))
        t.center.setValue((0, 2, 0))
        if invert:
            t.rotation.setValue(coin.SbVec3f((0, 0, 1)), -math.pi / 2)
        else:
            t.rotation.setValue(coin.SbVec3f((0, 0, 1)), math.pi / 2)
        c = coin.SoCone()
        c.height.setValue(4)
        marker.addChild(t)
        marker.addChild(c)
        return marker
    elif symbol == 3:
        marker = coin.SoSeparator()
        # hints are required otherwise only the bottom of the face is colored
        h = coin.SoShapeHints()
        h.vertexOrdering = h.COUNTERCLOCKWISE
        c = coin.SoCoordinate3()
        c.point.setValues([(-1, -2, 0), (0, 2, 0), (1, 2, 0), (0, -2, 0)])
        f = coin.SoFaceSet()
        marker.addChild(h)
        marker.addChild(c)
        marker.addChild(f)
        return marker
    elif symbol == 4:
        return dim_dash((-1.5, -1.5, 0), (1.5, 1.5, 0))
    elif symbol == 5:
        return coin.SoSeparator()
    else:
        _wrn(translate("draft", "Symbol not implemented. Using a default symbol."))
        return coin.SoSphere()


dimSymbol = dim_symbol


def dim_dash(p1, p2):
    """Return a SoSeparator with a line used to make dimension dashes.

    It is used by `dim_symbol` to create line end symbols
    like `'Tick-2'`, `'DimOvershoot'`, and `'ExtOvershoot'` dashes.

    Parameters
    ----------
    p1: tuple of three floats or Base::Vector3
        A point to define a line vertex.

    p2: tuple of three floats or Base::Vector3
        A point to define a line vertex.

    Returns
    -------
    Coin.SoSeparator
        A Coin object with a `SoLineSet` created from `p1` and `p2`
        as vertices.
    """
    dash = coin.SoSeparator()
    v = coin.SoVertexProperty()
    v.vertex.set1Value(0, p1)
    v.vertex.set1Value(1, p2)
    line = coin.SoLineSet()
    line.vertexProperty = v
    dash.addChild(line)
    return dash


dimDash = dim_dash


def remove_hidden(objectslist):
    """Return only the visible objects in the list.

    This function only works if the graphical interface is available
    as the `Visibility` attribute is a property of the view provider
    (`obj.ViewObject`).

    Parameters
    ----------
    objectslist: list of App::DocumentObject
        List of any type of object.

    Returns
    -------
    list
        Return a copy of the input list without those objects
        for which `obj.ViewObject.Visibility` is `False`.

        If the graphical interface is not loaded
        the returned list is just a copy of the input list.
    """
    newlist = objectslist[:]
    for obj in objectslist:
        if obj.ViewObject:
            if not obj.ViewObject.isVisible():
                newlist.remove(obj)
    return newlist


removeHidden = remove_hidden


def get_diffuse_color(objs):
    """Get a (cumulative) diffuse color from one or more objects.

    If all tuples in the result are identical a list with a single tuple is
    returned. In theory all faces of an object can be set to the same diffuse
    color that is different from its shape color, but that seems rare. The
    function does not take that into account.

    Parameters
    ----------
    objs: a single object or an iterable with objects.

    Returns
    -------
    list of tuples
        The list will be empty if no valid object is found.
    """

    def _get_color(obj):
        if hasattr(obj, "ColoredElements"):
            if hasattr(obj, "Count") or hasattr(obj, "ElementCount"):
                # Link and Link array.
                if hasattr(obj, "Count"):
                    count = obj.Count
                    base = obj.Base
                else:
                    count = obj.ElementCount if obj.ElementCount > 0 else 1
                    base = obj.LinkedObject
                if base is None:
                    return []
                cols = _get_color(base) * count
                if obj.ColoredElements is None:
                    return cols
                face_num = len(base.Shape.Faces)
                for elm, override in zip(obj.ColoredElements[1], obj.ViewObject.OverrideColorList):
                    if (
                        "Face" in elm
                    ):  # Examples: "Face3" and "1.Face6". Int before "." is zero-based, other int is 1-based.
                        if "." in elm:
                            elm0, elm1 = elm.split(".")
                            i = (int(elm0) * face_num) + int(elm1[4:]) - 1
                            cols[i] = override
                        else:
                            i = int(elm[4:]) - 1
                            for j in range(count):
                                cols[(j * face_num) + i] = override
                return cols
            elif hasattr(obj, "ElementList"):
                # LinkGroup
                cols = []
                for sub in obj.ElementList:
                    sub_cols = _get_color(sub)
                    if obj.ColoredElements is None:
                        cols += sub_cols
                    else:
                        for elm, override in zip(
                            obj.ColoredElements[1], obj.ViewObject.OverrideColorList
                        ):
                            if sub.Name + ".Face" in elm:
                                i = int(elm[(len(sub.Name) + 5) :]) - 1
                                sub_cols[i] = override
                        cols += sub_cols
                return cols
            else:
                return []
        elif hasattr(obj.ViewObject, "DiffuseColor"):
            if len(obj.ViewObject.DiffuseColor) == len(obj.Shape.Faces):
                return obj.ViewObject.DiffuseColor
            else:
                col = obj.ViewObject.ShapeColor
                col = (col[0], col[1], col[2], 1.0 - obj.ViewObject.Transparency / 100.0)
                return [col] * len(obj.Shape.Faces)
        elif obj.hasExtension("App::GeoFeatureGroupExtension"):
            cols = []
            for sub in obj.Group:
                cols += _get_color(sub)
            return cols
        else:
            return []

    if not isinstance(objs, list):
        # Quick check to avoid processing a single object:
        obj = objs
        if (
            not hasattr(obj, "ColoredElements")
            and hasattr(obj.ViewObject, "DiffuseColor")
            and (
                len(obj.ViewObject.DiffuseColor) == 1
                or len(obj.ViewObject.DiffuseColor) == len(obj.Shape.Faces)
            )
        ):
            return obj.ViewObject.DiffuseColor
        # Create a list for further processing:
        objs = [objs]

    colors = []
    for obj in objs:
        colors += _get_color(obj)

    if len(colors) > 1:
        first = colors[0]
        for next in colors[1:]:
            if next != first:
                break
        else:
            colors = [first]
    return colors


def apply_current_style(objs):
    """Apply the current style to one or more objects.

    Parameters
    ----------
    objs: a single object or an iterable with objects.
    """
    if not isinstance(objs, list):
        objs = [objs]
    anno_style = utils.get_default_annotation_style()
    shape_style = utils.get_default_shape_style()
    for obj in objs:
        if not hasattr(obj, "ViewObject"):
            continue
        vobj = obj.ViewObject
        props = vobj.PropertiesList
        style = anno_style if ("FontName" in props) else shape_style
        for prop in props:
            if prop in style:
                if style[prop][0] == "index":
                    if style[prop][2] in vobj.getEnumerationsOfProperty(prop):
                        setattr(vobj, prop, style[prop][2])
                else:
                    setattr(vobj, prop, style[prop][1])


def restore_view_object(obj, vp_module, vp_class, format=True, format_ref=None):
    """Restore the ViewObject if the object was saved without the GUI.

    Parameters
    ----------
    obj: App::DocumentObject
        Object whose ViewObject needs to be restored.

    vp_module: string
        View provider module. Must be in the draftviewproviders directory.

    vp_class: string
        View provider class.

    format: bool, optional
        Defaults to `True`.
        If `True` the `format_object` function is called to update the
        properties of the ViewObject.

    format_ref: App::DocumentObject, optional
        Defaults to `None`.
        Reference object to copy ViewObject properties from.
    """
    if not getattr(obj, "ViewObject", None):
        return
    vobj = obj.ViewObject
    if not getattr(vobj, "Proxy", None):
        vp_module = importlib.import_module("draftviewproviders." + vp_module)
        getattr(vp_module, vp_class)(vobj)
        if format:
            format_object(obj, format_ref)


def format_object(target, origin=None, ignore_construction=False):
    """Apply visual properties to an object.

    This function only works if the graphical interface is available.

    If origin is `None` and target is not an annotation, the DefaultDrawStyle
    and DefaultDisplayMode preferences are applied. Else, the properties of
    origin are applied to target.

    If construction mode is active target is then placed in the construction
    group and the `constr` color is applied to its applicable color properties:
    TextColor, PointColor, LineColor, and ShapeColor.

    Parameters
    ----------
    target: App::DocumentObject

    origin: App::DocumentObject, optional
        Defaults to `None`.
        If construction mode is not active, its visual properties are assigned
        to `target`, with the exception of `BoundingBox`, `Proxy`, `RootNode`
        and `Visibility`.

    ignore_construction: bool, optional
        Defaults to `False`.
        Set to `True` to ignore construction mode.
    """
    if not target:
        return
    if not App.GuiUp:
        return
    if not hasattr(Gui, "draftToolBar"):
        return
    if not hasattr(target, "ViewObject"):
        return
    if hasattr(target, "Shape") and target.Shape.Faces:
        len_faces = len(target.Shape.Faces)
    else:
        len_faces = 1
    obrep = target.ViewObject
    obprops = obrep.PropertiesList
    if origin and hasattr(origin, "ViewObject"):
        matchrep = origin.ViewObject
        for p in matchrep.PropertiesList:
            if p in ("DisplayMode", "BoundingBox", "Proxy", "RootNode", "Visibility"):
                continue
            if p not in obprops:
                continue
            if obrep.getEditorMode(p):
                continue
            val = getattr(matchrep, p)
            if isinstance(val, tuple):
                if len(val) != len_faces:
                    val = (val[0],)
            elif hasattr(val, "Value"):
                val = val.Value
            try:
                setattr(obrep, p, val)
            except Exception:
                pass
        if matchrep.DisplayMode in obrep.listDisplayModes():
            obrep.DisplayMode = matchrep.DisplayMode
        if hasattr(obrep, "DiffuseColor"):
            difcol = get_diffuse_color(origin)
            if difcol and len(difcol) == len_faces:
                obrep.DiffuseColor = difcol
    elif "FontName" not in obprops:
        # Apply 2 Draft style preferences, other style preferences are applied by Core.
        if "DrawStyle" in obprops:
            obrep.DrawStyle = utils.DRAW_STYLES[params.get_param("DefaultDrawStyle")]
        if "DisplayMode" in obprops:
            dm = utils.DISPLAY_MODES[params.get_param("DefaultDisplayMode")]
            if dm in obrep.listDisplayModes():
                obrep.DisplayMode = dm
    if ignore_construction:
        return
    if Gui.draftToolBar.isConstructionMode():
        doc = App.ActiveDocument
        col = params.get_param("constructioncolor") | 0x000000FF
        grp = doc.getObject("Draft_Construction")
        if not grp:
            grp = doc.addObject("App::DocumentObjectGroup", "Draft_Construction")
            grp.Label = params.get_param("constructiongroupname")
        grp.addObject(target)
        if "TextColor" in obprops:
            obrep.TextColor = col
        if "PointColor" in obprops:
            obrep.PointColor = col
        if "LineColor" in obprops:
            obrep.LineColor = col
        if "ShapeColor" in obprops:
            obrep.ShapeColor = col
        if hasattr(obrep, "Transparency"):
            obrep.Transparency = 80


formatObject = format_object


def get_selection(gui=App.GuiUp):
    """Return the current selected objects.

    This function only works if the graphical interface is available
    as the selection module only works on the 3D view.

    It wraps around `Gui.Selection.getSelection`

    Parameters
    ----------
    gui: bool, optional
        It defaults to the value of `App.GuiUp`, which is `True`
        when the interface exists, and `False` otherwise.

        This value can be set to `False` to simulate
        when the interface is not available.

    Returns
    -------
    list of App::DocumentObject
        Returns a list of objects in the current selection.
        It can be an empty list if no object is selected.

        If the interface is not available, it returns `None`.
    """
    if gui:
        return Gui.Selection.getSelection()
    return None


getSelection = get_selection


def get_selection_ex(gui=App.GuiUp):
    """Return the current selected objects together with their subelements.

    This function only works if the graphical interface is available
    as the selection module only works on the 3D view.

    It wraps around `Gui.Selection.getSelectionEx`

    Parameters
    ----------
    gui: bool, optional
        It defaults to the value of `App.GuiUp`, which is `True`
        when the interface exists, and `False` otherwise.

        This value can be set to `False` to simulate
        when the interface is not available.

    Returns
    -------
    list of Gui::SelectionObject
        Returns a list of `Gui::SelectionObject` in the current selection.
        It can be an empty list if no object is selected.

        If the interface is not available, it returns `None`.

    Selection objects
    -----------------
    One `Gui::SelectionObject` has attributes that indicate which specific
    subelements, that is, vertices, wires, and faces, were selected.
    This can be useful to operate on the subelements themselves.
    If `G` is a `Gui::SelectionObject`
     * `G.Object` is the selected object
     * `G.ObjectName` is the name of the selected object
     * `G.HasSubObjects` is `True` if there are subelements in the selection
     * `G.SubObjects` is a tuple of the subelements' shapes
     * `G.SubElementNames` is a tuple of the subelements' names

    `SubObjects` and `SubElementNames` should be empty tuples
    if `HasSubObjects` is `False`.
    """
    if gui:
        return Gui.Selection.getSelectionEx()
    return None


getSelectionEx = get_selection_ex


def select(objs=None, gui=App.GuiUp):
    """Unselects everything and selects only the given list of objects.

    This function only works if the graphical interface is available
    as the selection module only works on the 3D view.

    Parameters
    ----------
    objs: list of App::DocumentObjects or tuples, or a single object or tuple, optional
        It defaults to `None`.
        Format for tuples:
        `(doc.Name or "", sel.Object.Name, sel.SubElementName or "")`
        For example (Box nested in Part):
        `("", "Part", "Box.Edge1")`

    gui: bool, optional
        It defaults to the value of `App.GuiUp`, which is `True`
        when the interface exists, and `False` otherwise.

        This value can be set to `False` to simulate
        when the interface is not available.
    """
    if gui:
        Gui.Selection.clearSelection()
        if objs is not None:
            if not isinstance(objs, list):
                objs = [objs]
            for obj in objs:
                if not obj:
                    continue
                if isinstance(obj, tuple):
                    # Example of tuple (Rectangle in Part):
                    #   ("", "Part", "Rectangle.")
                    # See:
                    #   utils._modifiers_process_selection()
                    #   utils._modifiers_process_subselection()
                    parent = App.ActiveDocument.getObject(obj[1])
                    if parent and parent.getSubObject(obj[2]):
                        Gui.Selection.addSelection(*obj)
                    continue
                if utils.is_deleted(obj):
                    continue
                Gui.Selection.addSelection(obj)


def load_texture(filename, size=None, gui=App.GuiUp):
    """Return a Coin.SoSFImage to use as a texture for a 2D plane.

    This function only works if the graphical interface is available
    as the visual properties that can be applied to a shape
    are attributes of the view provider (`obj.ViewObject`).

    Parameters
    ----------
    filename: str
        A path to a pixel image file (PNG) that can be used as a texture
        on the face of an object.

    size: tuple of two int, or a single int, optional
        It defaults to `None`.
        If a tuple is given, the two values define the width and height
        in pixels to which the loaded image will be scaled.
        If it is a single value, it is used for both dimensions.

        If it is `None`, the size will be determined from the `QImage`
        created from `filename`.

        CURRENTLY the input `size` parameter IS NOT USED.
        It always uses the `QImage` to determine this information.

    gui: bool, optional
        It defaults to the value of `App.GuiUp`, which is `True`
        when the interface exists, and `False` otherwise.

        This value can be set to `False` to simulate
        when the interface is not available.

    Returns
    -------
    coin.SoSFImage
        An image object with the appropriate size, number of components
        (grayscale, grayscale and transparency, color,
        color and transparency), and byte data.

        It returns `None` if the interface is not available,
        or if there is a problem creating the image.
    """
    if gui:
        # from pivy import coin
        # from PySide import QtGui, QtSvg
        try:
            p = QtGui.QImage(filename)

            if p.isNull():
                _wrn("load_texture: " + translate("draft", "image is Null"))

                if not os.path.exists(filename):
                    raise FileNotFoundError(
                        -1,
                        translate(
                            "draft",
                            "filename does not exist " "on the system or " "in the resource file",
                        ),
                        filename,
                    )

            # This is buggy so it was de-activated.
            #
            # TODO: allow SVGs to use resolutions
            # if size and (".svg" in filename.lower()):
            #    # this is a pattern, not a texture
            #    if isinstance(size, int):
            #        size = (size, size)
            #    svgr = QtSvg.QSvgRenderer(filename)
            #    p = QtGui.QImage(size[0], size[1],
            #                     QtGui.QImage.Format_ARGB32)
            #    pa = QtGui.QPainter()
            #    pa.begin(p)
            #    svgr.render(pa)
            #    pa.end()
            # else:
            #    p = QtGui.QImage(filename)
            size = coin.SbVec2s(p.width(), p.height())
            buffersize = p.sizeInBytes()
            width = size[0]
            height = size[1]
            numcomponents = int(buffersize / (width * height))

            img = coin.SoSFImage()
            byteList = bytearray()

            # The SoSFImage needs to be filled with bytes.
            # The pixel information is converted into a Qt color, gray,
            # red, green, blue, or transparency (alpha),
            # depending on the input image.
            for y in range(height):
                # line = width*numcomponents*(height-(y));
                for x in range(width):
                    rgba = p.pixel(x, y)
                    if numcomponents <= 2:
                        byteList.append(QtGui.qGray(rgba))

                        if numcomponents == 2:
                            byteList.append(QtGui.qAlpha(rgba))

                    elif numcomponents <= 4:
                        byteList.append(QtGui.qRed(rgba))
                        byteList.append(QtGui.qGreen(rgba))
                        byteList.append(QtGui.qBlue(rgba))

                        if numcomponents == 4:
                            byteList.append(QtGui.qAlpha(rgba))
                    # line += numcomponents

            _bytes = bytes(byteList)
            img.setValue(size, numcomponents, _bytes)
        except FileNotFoundError as exc:
            _wrn("load_texture: {0}, {1}".format(exc.strerror, exc.filename))
            return None
        except Exception as exc:
            _wrn(str(exc))
            _wrn("load_texture: " + translate("draft", "unable to load texture"))
            return None
        else:
            return img
    return None


loadTexture = load_texture


def migrate_text_display_mode(obj_type="Text", mode="3D text", doc=None):
    """Migrate the display mode of objects of certain type."""
    if not doc:
        doc = App.activeDocument()

    for obj in doc.Objects:
        if utils.get_type(obj) == obj_type:
            obj.ViewObject.DisplayMode = mode


def get_bbox(obj, debug=False):
    """Return a BoundBox from any object that has a Coin RootNode.

    Normally the bounding box of an object can be taken
    from its `Part::TopoShape`.
    ::
        >>> print(obj.Shape.BoundBox)

    However, for objects without a `Shape`, such as those
    derived from `App::FeaturePython` like `Draft Text` and `Draft Dimension`,
    the bounding box can be calculated from the `RootNode` of the viewprovider.

    Parameters
    ----------
    obj: App::DocumentObject
        Any object that has a `ViewObject.RootNode`.

    Returns
    -------
    Base::BoundBox
        It returns a `BoundBox` object which has information like
        minimum and maximum values of X, Y, and Z, as well as bounding box
        center.

    None
        If there is a problem it will return `None`.
    """
    _name = "get_bbox"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    if isinstance(obj, str):
        obj_str = obj

    found, obj = utils.find_object(obj, doc)
    if not found:
        _err(translate("draft", "Wrong input: object {} not in document.").format(obj_str))
        return None

    if (
        not hasattr(obj, "ViewObject")
        or not obj.ViewObject
        or not hasattr(obj.ViewObject, "RootNode")
    ):
        _err(translate("draft", "Does not have 'ViewObject.RootNode'."))

    # For Draft Dimensions
    # node = obj.ViewObject.Proxy.node
    node = obj.ViewObject.RootNode

    view = Gui.ActiveDocument.ActiveView
    region = view.getViewer().getSoRenderManager().getViewportRegion()
    action = coin.SoGetBoundingBoxAction(region)

    node.getBoundingBox(action)
    bb = action.getBoundingBox()

    # xlength, ylength, zlength = bb.getSize().getValue()
    xmin, ymin, zmin = bb.getMin().getValue()
    xmax, ymax, zmax = bb.getMax().getValue()

    return App.BoundBox(xmin, ymin, zmin, xmax, ymax, zmax)


# Code by Yorik van Havre (adapted).
def find_coin_node(parent, nodetype):
    if not hasattr(parent, "getNumChildren"):
        return None
    for i in range(parent.getNumChildren()):
        if isinstance(parent.getChild(i), nodetype):
            return parent.getChild(i)
    return None


def find_coin_node_by_name(parent, name):
    if not hasattr(parent, "getNumChildren"):
        return None
    for i in range(parent.getNumChildren()):
        if parent.getChild(i).getName() == name:
            return parent.getChild(i)
    return None


# Code by Chris Hennes (chennes).
# See https://forum.freecad.org/viewtopic.php?p=656362#p656362.
# Used to fix https://github.com/FreeCAD/FreeCAD/issues/10469.
def end_all_events():
    view = get_3d_view()
    if view is None:
        return
    if view.getNavigationType() in (
        "Gui::GestureNavigationStyle",
        "Gui::MayaGestureNavigationStyle",
    ):
        return

    class DelayEnder:
        def __init__(self):
            self.delay_is_done = False

        def stop(self):
            self.delay_is_done = True

    ender = DelayEnder()
    timer = QtCore.QTimer()
    timer.timeout.connect(ender.stop)
    timer.setSingleShot(True)
    timer.start(
        100
    )  # 100ms (50ms is too short) timer guarantees the loop below runs at least that long
    while not ender.delay_is_done:
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)


def toggle_working_plane(obj, action=None, restore=False, dialog=None):
    """Toggle the active state of a working plane object.

    This function handles the common logic for activating and deactivating
    working plane objects like BuildingParts and WorkingPlaneProxies.
    It can be used by different modules that need to implement similar
    working plane activation behavior.

    Parameters
    ----------
    obj : App::DocumentObject
        The object to activate or deactivate as a working plane.
    action : QAction, optional
        The action button that triggered this function, to update its checked state.
    restore : bool, optional
        If True, will restore the previous working plane when deactivating.
        Defaults to False.
    dialog : QDialog, optional
        If provided, will update the checked state of the activate button in the dialog.

    Returns
    -------
    bool
        True if the object was activated, False if it was deactivated.
    """

    # Determine the appropriate context based on object type
    context = "Arch"
    obj_type = utils.get_type(obj)
    if obj_type == "IfcBuildingStorey":
        context = "NativeIFC"

    # Check if the object is already active in its context
    is_active_arch = Gui.ActiveDocument.ActiveView.getActiveObject("Arch") == obj
    is_active_ifc = Gui.ActiveDocument.ActiveView.getActiveObject("NativeIFC") == obj
    is_active = is_active_arch or is_active_ifc
    if is_active:
        # Deactivate the object
        if is_active_arch:
            Gui.ActiveDocument.ActiveView.setActiveObject("Arch", None)
        if is_active_ifc:
            Gui.ActiveDocument.ActiveView.setActiveObject("NativeIFC", None)

        if (
            hasattr(obj, "ViewObject")
            and hasattr(obj.ViewObject, "Proxy")
            and hasattr(obj.ViewObject.Proxy, "setWorkingPlane")
        ):
            obj.ViewObject.Proxy.setWorkingPlane(restore=True)
        if action:
            action.setChecked(False)
        if dialog and hasattr(dialog, "buttonActive"):
            dialog.buttonActive.setChecked(False)
        return False
    else:
        # Activate the object
        Gui.ActiveDocument.ActiveView.setActiveObject(context, obj)
        if (
            hasattr(obj, "ViewObject")
            and hasattr(obj.ViewObject, "Proxy")
            and hasattr(obj.ViewObject.Proxy, "setWorkingPlane")
        ):
            obj.ViewObject.Proxy.setWorkingPlane()
        if action:
            action.setChecked(True)
        if dialog and hasattr(dialog, "buttonActive"):
            dialog.buttonActive.setChecked(True)
        return True


## @}



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint initialtemperature"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

from FreeCAD import Units


def get_analysis_types():
    return ["thermomech", "static"]


def get_sets_name():
    return "constraints_initial_temperature_node_sets"


def get_constraint_title():
    return "Initial temperature constraint"


def write_meshdata_constraint(f, femobj, inittemp_obj, ccxwriter):
    if inittemp_obj.References and len(inittemp_obj.References) > 0:
        f.write(f"*NSET,NSET={inittemp_obj.Name}\n")
        for n in femobj["Nodes"]:
            f.write(f"{n},\n")
    else:
        return


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return "*INITIAL CONDITIONS,TYPE=TEMPERATURE\n"


def get_after_write_constraint():
    return ""


def write_constraint(f, femobj, inittemp_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    initialtemp = inittemp_obj.initialTemperature.getValueAs("K")

    if inittemp_obj.References and len(inittemp_obj.References) > 0:
        f.write(f"{inittemp_obj.Name},{initialtemp}\n")
    else:
        f.write(f"{ccxwriter.ccx_nall},{initialtemp}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2023 Werner Mayer LGPL

__title__="Document handling module"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Tools for extracting or creating project files"


import os
import xml.sax
import xml.sax.handler
import xml.sax.xmlreader
import zipfile

# SAX handler to parse the Document.xml
class DocumentHandler(xml.sax.handler.ContentHandler):
    """ Parse content of Document.xml or GuiDocument.xml """
    def __init__(self, dirname):
        """ Init parser """
        super().__init__()
        self.files = []
        self.dirname = dirname

    def startElement(self, name, attrs):
        item = attrs.get("file")
        if item is not None:
            self.files.append(os.path.join(self.dirname, str(item)))

    def characters(self, content):
        return

    def endElement(self, name):
        return

def extractDocument(filename, outpath):
    """ Extract files from project archive """
    zfile = zipfile.ZipFile(filename)
    files = zfile.namelist()

    for i in files:
        data = zfile.read(i)
        dirs = i.split("/")
        if len(dirs) > 1:
            dirs.pop()
            curpath = outpath
            for j in dirs:
                curpath = curpath + "/" + j
                os.mkdir(curpath)
        output = open(outpath + "/"+i, 'wb')
        output.write(data)
        output.close()

def createDocument(filename, outpath):
    """ Create project archive """
    files = getFilesList(filename)
    dirname = os.path.dirname(filename)
    guixml = os.path.join(dirname, "GuiDocument.xml")
    if os.path.exists(guixml):
        files.extend(getFilesList(guixml))
    compress = zipfile.ZipFile(outpath, 'w', zipfile.ZIP_DEFLATED)
    for file in files:
        if os.path.isfile(file):
            path_in_archive = os.path.relpath(path=file, start=dirname)
            compress.write(file, path_in_archive, zipfile.ZIP_DEFLATED)
    compress.close()

def getFilesList(filename):
    """ Determine list of files referenced in a Document.xml or GuiDocument.xml """
    dirname = os.path.dirname(filename)
    handler = DocumentHandler(dirname)
    parser = xml.sax.make_parser()
    parser.setContentHandler(handler)
    parser.parse(filename)

    files = []
    files.append(filename)
    files.extend(iter(handler.files))
    return files



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import logging
import asyncio
import threading
import pathlib
import hashlib
from typing import (
    Dict,
    Any,
    Type,
    Optional,
    List,
    Sequence,
    Union,
    Set,
    Mapping,
    Tuple,
)
from dataclasses import dataclass
from PySide import QtCore, QtGui
from .store.base import AssetStore
from .asset import Asset
from .serializer import AssetSerializer
from .uri import AssetUri
from .cache import AssetCache, CacheKey


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.ERROR)


@dataclass
class _AssetConstructionData:
    """Holds raw data and type info needed to construct an asset instance."""

    store: str  # Name of the store where this asset was found
    uri: AssetUri
    raw_data: bytes
    asset_class: Type[Asset]
    # Stores AssetConstructionData for dependencies, keyed by their AssetUri
    dependencies_data: Optional[Dict[AssetUri, Optional["_AssetConstructionData"]]] = None


class AssetManager:
    def __init__(self, cache_max_size_bytes: int = 100 * 1024 * 1024):
        self.stores: Dict[str, AssetStore] = {}
        self._serializers: List[Tuple[Type[AssetSerializer], Type[Asset]]] = []
        self._asset_classes: Dict[str, Type[Asset]] = {}
        self.asset_cache = AssetCache(max_size_bytes=cache_max_size_bytes)
        self._cacheable_stores: Set[str] = set()
        logger.debug(f"AssetManager initialized (Thread: {threading.current_thread().name})")

    def register_store(self, store: AssetStore, cacheable: bool = False):
        """Registers an AssetStore with the manager."""
        logger.debug(f"Registering store: {store.name}, cacheable: {cacheable}")
        self.stores[store.name] = store
        if cacheable:
            self._cacheable_stores.add(store.name)

    def get_serializer_for_class(self, asset_class: Type[Asset]):
        for serializer, theasset_class in self._serializers:
            if issubclass(asset_class, theasset_class):
                return serializer
        raise ValueError(f"No serializer found for class {asset_class}")

    def register_asset(self, asset_class: Type[Asset], serializer: Type[AssetSerializer]):
        """Registers an Asset class with the manager."""
        if not issubclass(asset_class, Asset):
            raise TypeError(f"Item '{asset_class.__name__}' must be a subclass of Asset.")
        if not issubclass(serializer, AssetSerializer):
            raise TypeError(f"Item '{serializer.__name__}' must be a subclass of AssetSerializer.")
        self._serializers.append((serializer, asset_class))

        asset_type_name = getattr(asset_class, "asset_type", None)
        if not isinstance(asset_type_name, str) or not asset_type_name:  # Ensure not empty
            raise TypeError(
                f"Asset class '{asset_class.__name__}' must have a non-empty string 'asset_type' attribute."
            )

        logger.debug(f"Registering asset type: '{asset_type_name}' -> {asset_class.__name__}")
        self._asset_classes[asset_type_name] = asset_class

    async def _fetch_asset_construction_data_recursive_async(
        self,
        uri: AssetUri,
        store_names: Sequence[str],
        visited_uris: Set[AssetUri],
        depth: Optional[int] = None,
    ) -> Optional[_AssetConstructionData]:
        # Log library fetch details
        if uri.asset_type == "library":
            logger.info(
                f"LIBRARY FETCH: Loading library '{uri.asset_id}' with depth={depth} from stores {store_names}"
            )

        logger.debug(
            f"_fetch_asset_construction_data_recursive_async called {store_names} {uri} {depth}"
        )

        if uri in visited_uris:
            logger.error(f"Cyclic dependency detected for URI: {uri}")
            raise RuntimeError(f"Cyclic dependency encountered for URI: {uri}")

        # Check arguments
        if not store_names:
            raise ValueError("At least one store name must be provided.")

        asset_class = self._asset_classes.get(uri.asset_type)
        if not asset_class:
            raise ValueError(f"No asset class registered for URI: {uri}")

        # Fetch the requested asset, trying each store in order
        raw_data = None
        found_store_name = None

        # Log toolbit search details
        if uri.asset_type == "toolbit":
            logger.info(
                f"TOOLBIT SEARCH: Looking for toolbit '{uri.asset_id}' in stores: {store_names}"
            )

        for current_store_name in store_names:
            store = self.stores.get(current_store_name)
            if not store:
                logger.warning(f"Store '{current_store_name}' not registered. Skipping.")
                continue

            # Log store search path for toolbits
            if uri.asset_type == "toolbit":
                store_path = getattr(store, "base_path", "unknown")
                logger.info(
                    f"TOOLBIT SEARCH: Checking store '{current_store_name}' at path: {store_path}"
                )

            try:
                raw_data = await store.get(uri)
                found_store_name = current_store_name
                if uri.asset_type == "toolbit":
                    logger.info(
                        f"TOOLBIT FOUND: '{uri.asset_id}' found in store '{found_store_name}'"
                    )
                logger.debug(
                    f"_fetch_asset_construction_data_recursive_async: Asset {uri} found in store {found_store_name}"
                )
                break  # Asset found, no need to check other stores
            except FileNotFoundError:
                if uri.asset_type == "toolbit":
                    logger.info(
                        f"TOOLBIT SEARCH: '{uri.asset_id}' NOT found in store '{current_store_name}'"
                    )
                logger.debug(
                    f"_fetch_asset_construction_data_recursive_async: Asset {uri} not found in store {current_store_name}"
                )
                continue  # Try next store

        if raw_data is None or not found_store_name:
            if uri.asset_type == "toolbit":
                logger.warning(
                    f"TOOLBIT NOT FOUND: '{uri.asset_id}' not found in any of the stores: {store_names}"
                )
            return None  # Asset not found in any store

        if depth == 0:
            if uri.asset_type == "library":
                logger.warning(
                    f"LIBRARY SHALLOW: Library '{uri.asset_id}' loaded with depth=0 - no dependencies will be resolved"
                )
            return _AssetConstructionData(
                store=found_store_name,
                uri=uri,
                raw_data=raw_data,
                asset_class=asset_class,
                dependencies_data=None,  # Indicates that no attempt was made to fetch deps
            )

        # Extract the list of dependencies (non-recursive)
        serializer = self.get_serializer_for_class(asset_class)
        dependency_uris = asset_class.extract_dependencies(raw_data, serializer)

        # Initialize deps_construction_data_map. Any dependencies mapped to None
        # indicate that dependencies were intentionally not fetched.
        deps_construction_data: Dict[AssetUri, Optional[_AssetConstructionData]] = {}

        for dep_uri in dependency_uris:
            visited_uris.add(uri)
            try:
                # For dependencies, use the same list of stores for fallback
                dep_data = await self._fetch_asset_construction_data_recursive_async(
                    dep_uri,
                    store_names,  # Pass the list of stores for dependency fallback
                    visited_uris,
                    None if depth is None else depth - 1,
                )
            finally:
                visited_uris.remove(uri)
            deps_construction_data[dep_uri] = dep_data

        logger.debug(
            f"ToolBitShape '{uri.asset_id}' dependencies_data: {deps_construction_data is None}"
        )

        return _AssetConstructionData(
            store=found_store_name,
            uri=uri,
            raw_data=raw_data,
            asset_class=asset_class,
            dependencies_data=deps_construction_data,  # Can be None or Dict
        )

    def _calculate_cache_key_from_construction_data(
        self,
        construction_data: _AssetConstructionData,
    ) -> Optional[CacheKey]:
        if not construction_data or not construction_data.raw_data:
            return None

        if construction_data.dependencies_data is None:
            deps_signature_tuple: Tuple = ("shallow_children",)
        else:
            deps_signature_tuple = tuple(
                sorted(str(uri) for uri in construction_data.dependencies_data.keys())
            )

        raw_data_hash = int(hashlib.sha256(construction_data.raw_data).hexdigest(), 16)

        return CacheKey(
            store_name=construction_data.store,
            asset_uri_str=str(construction_data.uri),
            raw_data_hash=raw_data_hash,
            dependency_signature=deps_signature_tuple,
        )

    def _build_asset_tree_from_data_sync(
        self,
        construction_data: Optional[_AssetConstructionData],
    ) -> Optional[Asset]:
        """
        Synchronously and recursively builds an asset instance.
        Integrates caching logic.
        """
        if not construction_data:
            return None

        cache_key: Optional[CacheKey] = None
        if construction_data.store in self._cacheable_stores:
            cache_key = self._calculate_cache_key_from_construction_data(construction_data)
            if cache_key:
                cached_asset = self.asset_cache.get(cache_key)
                if cached_asset is not None:
                    return cached_asset

        logger.debug(
            f"BuildAssetTreeSync: Instantiating '{construction_data.uri}' "
            f"of type '{construction_data.asset_class.__name__}'"
        )

        resolved_dependencies: Optional[Mapping[AssetUri, Any]] = None
        if construction_data.dependencies_data is not None:
            resolved_dependencies = {}

            # Log dependency resolution for libraries
            if construction_data.uri.asset_type == "library":
                logger.info(
                    f"LIBRARY DEPS: Resolving {len(construction_data.dependencies_data)} dependencies for library '{construction_data.uri.asset_id}'"
                )

            for (
                dep_uri,
                dep_data_node,
            ) in construction_data.dependencies_data.items():
                # Log toolbit dependency resolution
                if dep_uri.asset_type == "toolbit":
                    logger.info(
                        f"TOOLBIT DEP: Resolving dependency '{dep_uri.asset_id}' for library '{construction_data.uri.asset_id}'"
                    )

                # Assuming dependencies are fetched from the same store context
                # for caching purposes. If a dependency *could* be from a
                # different store and that store has different cacheability,
                # this would need more complex store_name propagation.
                # For now, use the parent's store_name_for_cache.
                try:
                    dep = self._build_asset_tree_from_data_sync(dep_data_node)
                    if dep_uri.asset_type == "toolbit":
                        if dep:
                            logger.info(
                                f"TOOLBIT DEP: Successfully resolved '{dep_uri.asset_id}' -> {type(dep).__name__}"
                            )
                        else:
                            logger.warning(
                                f"TOOLBIT DEP: Dependency '{dep_uri.asset_id}' resolved to None"
                            )
                except Exception as e:
                    if dep_uri.asset_type == "toolbit":
                        logger.error(f"TOOLBIT DEP: Error resolving '{dep_uri.asset_id}': {e}")
                    logger.error(
                        f"Error building dependency '{dep_uri}' for asset '{construction_data.uri}': {e}",
                        exc_info=True,
                    )
                else:
                    resolved_dependencies[dep_uri] = dep

            # Log final dependency count for libraries
            if construction_data.uri.asset_type == "library":
                toolbit_deps = [
                    uri for uri in resolved_dependencies.keys() if uri.asset_type == "toolbit"
                ]
                logger.info(
                    f"LIBRARY DEPS: Resolved {len(resolved_dependencies)} total dependencies ({len(toolbit_deps)} toolbits) for library '{construction_data.uri.asset_id}'"
                )
        else:
            # Log when dependencies_data is None
            if construction_data.uri.asset_type == "library":
                logger.warning(
                    f"LIBRARY NO DEPS: Library '{construction_data.uri.asset_id}' has dependencies_data=None - was loaded with depth=0"
                )

        asset_class = construction_data.asset_class
        serializer = self.get_serializer_for_class(asset_class)
        try:
            # Log library instantiation with dependency info
            if construction_data.uri.asset_type == "library":
                dep_count = len(resolved_dependencies) if resolved_dependencies else 0
                logger.info(
                    f"LIBRARY INSTANTIATE: Creating library '{construction_data.uri.asset_id}' with {dep_count} dependencies"
                )

            final_asset = asset_class.from_bytes(
                construction_data.raw_data,
                construction_data.uri.asset_id,
                resolved_dependencies,
                serializer,
            )
        except Exception as e:
            logger.error(
                f"Error instantiating asset '{construction_data.uri}' of type '{asset_class.__name__}': {e}",
                exc_info=True,
            )
            return None

        if final_asset is not None and cache_key:
            # This check implies store_name_for_cache was in _cacheable_stores
            direct_deps_uris_strs: Set[str] = set()
            if construction_data.dependencies_data is not None:
                direct_deps_uris_strs = {
                    str(uri) for uri in construction_data.dependencies_data.keys()
                }
            raw_data_size = len(construction_data.raw_data)
            self.asset_cache.put(
                cache_key,
                final_asset,
                raw_data_size,
                direct_deps_uris_strs,
            )
        return final_asset

    def get(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> Asset:
        """
        Retrieves an asset by its URI (synchronous wrapper), to a specified depth.
        IMPORTANT: Assumes this method is CALLED ONLY from the main UI thread
                   if Asset.from_bytes performs UI operations.
        Depth None means infinite depth. Depth 0 means only this asset, no dependencies.
        """
        # Log entry with thread info for verification
        calling_thread_name = threading.current_thread().name
        stores_list = [store] if isinstance(store, str) else store

        # Log all asset get requests
        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri
        if asset_uri_obj.asset_type == "library":
            logger.info(
                f"LIBRARY GET: Request for library '{asset_uri_obj.asset_id}' with depth={depth}"
            )
        elif asset_uri_obj.asset_type == "toolbit":
            logger.info(
                f"TOOLBIT GET: Direct request for toolbit '{asset_uri_obj.asset_id}' with depth={depth} from stores {stores_list}"
            )
            # Add stack trace to see who's calling this
            import traceback

            stack = traceback.format_stack()
            caller_info = "".join(stack[-3:-1])  # Get the 2 frames before this one
            logger.info(f"TOOLBIT GET CALLER:\n{caller_info}")

        logger.debug(
            f"AssetManager.get(uri='{uri}', stores='{stores_list}', depth='{depth}') called from thread: {calling_thread_name}"
        )
        if (
            QtGui.QApplication.instance()
            and QtCore.QThread.currentThread() is not QtGui.QApplication.instance().thread()
        ):
            logger.warning(
                "AssetManager.get() called from a non-main thread! UI in from_bytes may fail!"
            )

        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri

        # Step 1: Fetch all data using asyncio.run
        try:
            logger.debug(
                f"Get: Starting asyncio.run for data fetching of '{asset_uri_obj}', depth {depth}."
            )
            all_construction_data = asyncio.run(
                self._fetch_asset_construction_data_recursive_async(
                    asset_uri_obj, stores_list, set(), depth
                )
            )
            logger.debug(
                f"Get: asyncio.run for data fetching of '{asset_uri_obj}', depth {depth} completed."
            )
        except Exception as e:
            logger.error(
                f"Get: Error during asyncio.run data fetching for '{asset_uri_obj}': {e}",
                exc_info=False,
            )
            raise  # Re-raise the exception from the async part

        if all_construction_data is None:
            # This means the top-level asset itself was not found by _fetch_...
            raise FileNotFoundError(f"Asset '{asset_uri_obj}' not found in stores '{stores_list}'")

        # Step 2: Synchronously build the asset tree (and call from_bytes)
        # This happens in the current thread (which is assumed to be the main UI thread)
        deps_count = 0
        found_deps_count = 0
        if all_construction_data.dependencies_data is not None:
            deps_count = len(all_construction_data.dependencies_data)
            found_deps_count = sum(
                1
                for d in all_construction_data.dependencies_data.values()
                if d is not None  # Count actual data, not None placeholders
            )

        logger.debug(
            f"Get: Starting synchronous asset tree build for '{asset_uri_obj}' "
            f"and {deps_count} dependencies ({found_deps_count} resolved)."
        )
        # Use the first store from the list for caching purposes
        final_asset = self._build_asset_tree_from_data_sync(all_construction_data)
        if not final_asset:
            raise ValueError(f"failed to build asset {uri}")
        logger.debug(f"Get: Synchronous asset tree build for '{asset_uri_obj}' completed.")
        return final_asset

    def get_or_none(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> Optional[Asset]:
        """
        Convenience wrapper for get() that does not raise FileNotFoundError; returns
        None instead
        """
        try:
            return self.get(uri, store, depth)
        except FileNotFoundError:
            return None

    async def get_async(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> Optional[Asset]:
        """
        Retrieves an asset by its URI (asynchronous), to a specified depth.
        NOTE: If Asset.from_bytes does UI work, this method should ideally be awaited
        from an asyncio loop that is integrated with the main UI thread (e.g., via QtAsyncio).
        If awaited from a plain worker thread's asyncio loop, from_bytes will run on that worker.
        """
        calling_thread_name = threading.current_thread().name
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"AssetManager.get_async(uri='{uri}', stores='{stores_list}', depth='{depth}') called from thread: {calling_thread_name}"
        )

        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri

        all_construction_data = await self._fetch_asset_construction_data_recursive_async(
            asset_uri_obj, stores_list, set(), depth
        )

        if all_construction_data is None:
            # Consistent with get(), if the top-level asset is not found,
            # raise FileNotFoundError.
            raise FileNotFoundError(
                f"Asset '{asset_uri_obj}' not found in stores '{stores_list}' (async path)"
            )
            # return None # Alternative: if Optional[Asset] means asset might not exist

        # Instantiation happens in the context of where this get_async was awaited.
        logger.debug(
            f"get_async: Building asset tree for '{asset_uri_obj}', depth {depth} in current async context."
        )
        return self._build_asset_tree_from_data_sync(all_construction_data)

    def get_raw(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
    ) -> bytes:
        """Retrieves raw asset data by its URI (synchronous wrapper)."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"AssetManager.get_raw(uri='{uri}', stores='{stores_list}') from T:{threading.current_thread().name}"
        )

        try:
            return asyncio.run(self.get_raw_async(uri, stores_list))
        except Exception as e:
            logger.error(
                f"GetRaw: Error during asyncio.run for '{uri}': {e}",
                exc_info=False,
            )
            raise

    async def get_raw_async(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
    ) -> bytes:
        """Retrieves raw asset data by its URI (asynchronous)."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"AssetManager.get_raw_async(uri='{uri}', stores='{stores_list}') from T:{threading.current_thread().name}"
        )
        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri

        for current_store_name in stores_list:
            thestore = self.stores.get(current_store_name)
            if not thestore:
                logger.warning(f"Store '{current_store_name}' not registered. Skipping.")
                continue
            try:
                raw_data = await thestore.get(asset_uri_obj)
                logger.debug(
                    f"GetRawAsync: Asset {asset_uri_obj} found in store {current_store_name}"
                )
                return raw_data
            except FileNotFoundError:
                logger.debug(
                    f"GetRawAsync: Asset {asset_uri_obj} not found in store {current_store_name}"
                )
                continue

        raise FileNotFoundError(f"Asset '{asset_uri_obj}' not found in stores '{stores_list}'")

    def get_bulk(
        self,
        uris: Sequence[Union[AssetUri, str]],
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> List[Any]:
        """Retrieves multiple assets by their URIs (synchronous wrapper), to a specified depth."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"AssetManager.get_bulk for {len(uris)} URIs from stores '{stores_list}', depth '{depth}'"
        )

        async def _fetch_all_construction_data_bulk_async():
            tasks = [
                self._fetch_asset_construction_data_recursive_async(
                    AssetUri(u) if isinstance(u, str) else u,
                    stores_list,
                    set(),
                    depth,
                )
                for u in uris
            ]
            # Gather all construction data concurrently
            # return_exceptions=True means results list can contain exceptions
            return await asyncio.gather(*tasks, return_exceptions=True)

        try:
            logger.debug("GetBulk: Starting bulk data fetching")
            all_construction_data_list = asyncio.run(_fetch_all_construction_data_bulk_async())
            logger.debug("GetBulk: bulk data fetching completed")
        except Exception as e:  # Should ideally not happen if gather returns exceptions
            logger.error(
                f"GetBulk: Unexpected error during asyncio.run for bulk data: {e}",
                exc_info=False,
            )
            raise

        assets = []
        for i, data_or_exc in enumerate(all_construction_data_list):
            original_uri_input = uris[i]
            # Explicitly re-raise exceptions found in the results list
            if isinstance(data_or_exc, Exception):
                logger.error(
                    f"GetBulk: Re-raising exception for '{original_uri_input}': {data_or_exc}",
                    exc_info=False,
                )
                raise data_or_exc
            elif isinstance(data_or_exc, _AssetConstructionData):
                # Build asset instance synchronously. Exceptions during build should propagate.
                # Use the first store from the list for caching purposes in build_asset_tree
                assets.append(self._build_asset_tree_from_data_sync(data_or_exc))
            elif data_or_exc is None:  # From _fetch_... returning None for not found
                logger.debug(f"GetBulk: Asset '{original_uri_input}' not found")
                assets.append(None)
            else:  # Should not happen
                logger.error(
                    f"GetBulk: Unexpected item in construction data list for '{original_uri_input}': {type(data_or_exc)}"
                )
                # Raise an exception for unexpected data types
                raise RuntimeError(
                    f"Unexpected data type for {original_uri_input}: {type(data_or_exc)}"
                )
        return assets

    async def get_bulk_async(
        self,
        uris: Sequence[Union[AssetUri, str]],
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> List[Any]:
        """Retrieves multiple assets by their URIs (asynchronous), to a specified depth."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"AssetManager.get_bulk_async for {len(uris)} URIs from stores '{stores_list}', depth '{depth}'"
        )
        tasks = [
            self._fetch_asset_construction_data_recursive_async(
                AssetUri(u) if isinstance(u, str) else u,
                stores_list,
                set(),
                depth,
            )
            for u in uris
        ]
        all_construction_data_list = await asyncio.gather(*tasks, return_exceptions=True)

        assets = []
        for i, data_or_exc in enumerate(all_construction_data_list):
            if isinstance(data_or_exc, _AssetConstructionData):
                # Use the first store from the list for caching purposes in build_asset_tree
                asset = self._build_asset_tree_from_data_sync(data_or_exc)
                assets.append(asset)
            elif isinstance(data_or_exc, FileNotFoundError) or data_or_exc is None:
                assets.append(None)
            elif isinstance(data_or_exc, Exception):
                assets.append(data_or_exc)  # Caller must check
        return assets

    def exists(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
    ) -> bool:
        """
        Returns True if the asset exists in any of the specified stores, False otherwise.
        """

        async def _exists_async(stores_list: Sequence[str]):
            asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri
            logger.debug(
                f"ExistsAsync (internal): Trying stores '{stores_list}'. Available stores: {list(self.stores.keys())}"
            )
            for current_store_name in stores_list:
                store = self.stores.get(current_store_name)
                if not store:
                    logger.error(f"Store '{current_store_name}' not registered. Skipping.")
                    raise ValueError(f"No store registered for name: {store}")
                try:
                    exists = await store.exists(asset_uri_obj)
                    if exists:
                        logger.debug(
                            f"ExistsAsync: Asset {asset_uri_obj} found in store {current_store_name}"
                        )
                        return True
                    else:
                        logger.debug(
                            f"ExistsAsync: Asset {asset_uri_obj} not found in store {current_store_name}"
                        )
                        continue
                except Exception as e:
                    logger.error(
                        f"ExistsAsync: Error checking store '{current_store_name}': {e}",
                        exc_info=False,
                    )
                    continue  # Try next store
            return False  # Not found in any store

        stores_list = [store] if isinstance(store, str) else store
        try:
            return asyncio.run(_exists_async(stores_list))
        except Exception as e:
            logger.error(
                f"AssetManager.exists: Error during asyncio.run for '{uri}': {e}",
                exc_info=False,
            )
            raise

    def fetch(
        self,
        asset_type: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> List[Asset]:
        """Fetches asset instances based on type, limit, and offset (synchronous), to a specified depth."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"Fetch(type='{asset_type}', stores='{stores_list}', depth='{depth}')")
        # Note: list_assets currently only supports a single store.
        # If fetching from multiple stores is needed for listing, this needs
        # to be updated. For now, list from the first store.
        list_store = stores_list[0] if stores_list else "local"
        asset_uris = self.list_assets(asset_type, limit, offset, list_store)
        results = self.get_bulk(asset_uris, stores_list, depth)  # Pass stores_list to get_bulk
        # Filter out non-Asset objects (e.g., None for not found, or exceptions if collected)
        return [asset for asset in results if isinstance(asset, Asset)]

    async def fetch_async(
        self,
        asset_type: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        store: Union[str, Sequence[str]] = "local",
        depth: Optional[int] = None,
    ) -> List[Asset]:
        """Fetches asset instances based on type, limit, and offset (asynchronous), to a specified depth."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"FetchAsync(type='{asset_type}', stores='{stores_list}', depth='{depth}')")
        # Note: list_assets_async currently only supports a single store.
        # If fetching from multiple stores is needed for listing, this needs
        # to be updated. For now, list from the first store.
        list_store = stores_list[0] if stores_list else "local"
        asset_uris = await self.list_assets_async(asset_type, limit, offset, list_store)
        results = await self.get_bulk_async(
            asset_uris, stores_list, depth
        )  # Pass stores_list to get_bulk_async
        return [asset for asset in results if isinstance(asset, Asset)]

    def list_assets(
        self,
        asset_type: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        store: Union[str, Sequence[str]] = "local",
    ) -> List[AssetUri]:
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"ListAssets(type='{asset_type}', stores='{stores_list}')")
        # Note: list_assets_async currently only supports a single store.
        # If listing from multiple stores is needed, this needs to be updated.
        # For now, list from the first store.
        list_store = stores_list[0] if stores_list else "local"
        return asyncio.run(self.list_assets_async(asset_type, limit, offset, list_store))

    async def list_assets_async(
        self,
        asset_type: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        store: Union[str, Sequence[str]] = "local",
    ) -> List[AssetUri]:
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"ListAssetsAsync executing for type='{asset_type}', stores='{stores_list}'")
        # Note: list_assets_async currently only supports a single store.
        # If listing from multiple stores is needed, this needs to be updated.
        # For now, list from the first store.
        list_store = stores_list[0] if stores_list else "local"
        logger.debug(
            f"ListAssetsAsync: Looking up store '{list_store}'. Available stores: {list(self.stores.keys())}"
        )
        try:
            selected_store = self.stores[list_store]
        except KeyError:
            raise ValueError(f"No store registered for name: {list_store}")
        return await selected_store.list_assets(asset_type, limit, offset)

    def count_assets(
        self,
        asset_type: Optional[str] = None,
        store: Union[str, Sequence[str]] = "local",
    ) -> int:
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"CountAssets(type='{asset_type}', stores='{stores_list}')")
        # Note: count_assets_async currently only supports a single store.
        # If counting across multiple stores is needed, this needs to be updated.
        # For now, count from the first store.
        count_store = stores_list[0] if stores_list else "local"
        return asyncio.run(self.count_assets_async(asset_type, count_store))

    async def count_assets_async(
        self,
        asset_type: Optional[str] = None,
        store: Union[str, Sequence[str]] = "local",
    ) -> int:
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"CountAssetsAsync executing for type='{asset_type}', stores='{stores_list}'")
        # Note: count_assets_async currently only supports a single store.
        # If counting across multiple stores is needed, this needs to be updated.
        # For now, count from the first store.
        count_store = stores_list[0] if stores_list else "local"
        logger.debug(
            f"CountAssetsAsync: Looking up store '{count_store}'. Available stores: {list(self.stores.keys())}"
        )
        try:
            selected_store = self.stores[count_store]
        except KeyError:
            raise ValueError(f"No store registered for name: {count_store}")
        return await selected_store.count_assets(asset_type)

    def _is_registered_type(self, obj: Asset) -> bool:
        """Helper to extract asset_type, id, and data from an object instance."""
        for registered_class_type in self._asset_classes.values():
            if isinstance(obj, registered_class_type):
                return True
        return False

    async def add_async(self, obj: Asset, store: str = "local") -> AssetUri:
        """
        Adds an asset to the store, either creating a new one or updating an existing one.
        Uses obj.get_url() to determine if the asset exists.
        """
        logger.debug(f"AddAsync: Adding {type(obj).__name__} to store '{store}'")
        uri = obj.get_uri()
        if not self._is_registered_type(obj):
            logger.warning(f"Asset has unregistered type '{uri.asset_type}' ({type(obj).__name__})")

        serializer = self.get_serializer_for_class(obj.__class__)
        data = obj.to_bytes(serializer)
        return await self.add_raw_async(uri.asset_type, uri.asset_id, data, store)

    def add(self, obj: Asset, store: str = "local") -> AssetUri:
        """Synchronous wrapper for adding an asset to the store."""
        logger.debug(
            f"Add: Adding {type(obj).__name__} to store '{store}' from T:{threading.current_thread().name}"
        )
        return asyncio.run(self.add_async(obj, store))

    async def add_raw_async(
        self, asset_type: str, asset_id: str, data: bytes, store: str = "local"
    ) -> AssetUri:
        """
        Adds raw asset data to the store, either creating a new asset or updating an existing one.
        """
        logger.debug(f"AddRawAsync: type='{asset_type}', id='{asset_id}', store='{store}'")
        if not asset_type or not asset_id:
            raise ValueError("asset_type and asset_id must be provided for add_raw.")
        if not isinstance(data, bytes):
            raise TypeError("Data for add_raw must be bytes.")
        selected_store = self.stores.get(store)
        if not selected_store:
            raise ValueError(f"No store registered for name: {store}")
        uri = AssetUri.build(asset_type=asset_type, asset_id=asset_id)
        try:
            uri = await selected_store.update(uri, data)
            logger.debug(f"AddRawAsync: Updated existing asset at {uri}")
        except FileNotFoundError:
            logger.debug(
                f"AddRawAsync: Asset not found, creating new asset with {asset_type} and {asset_id}"
            )
            uri = await selected_store.create(asset_type, asset_id, data)

        if store in self._cacheable_stores:
            self.asset_cache.invalidate_for_uri(str(uri))  # Invalidate after add/update
        return uri

    def add_raw(
        self, asset_type: str, asset_id: str, data: bytes, store: str = "local"
    ) -> AssetUri:
        """Synchronous wrapper for adding raw asset data to the store."""
        logger.debug(
            f"AddRaw: type='{asset_type}', id='{asset_id}', store='{store}' from T:{threading.current_thread().name}"
        )
        try:
            return asyncio.run(self.add_raw_async(asset_type, asset_id, data, store))
        except Exception as e:
            logger.error(
                f"AddRaw: Error for type='{asset_type}', id='{asset_id}': {e}",
                exc_info=False,
            )
            raise

    async def copy_async(
        self,
        src: AssetUri,
        dest_store: str,
        store: str = "local",
        dest: Optional[AssetUri] = None,
    ) -> AssetUri:
        """
        Copies an asset from one location to another asynchronously.

        Performs a shallow copy by wrapping get_raw_async and add_raw_async.
        If dest is None, it defaults to the uri given in src.
        An assertion is raised if src and store are the same as dest and
        dest_store.
        If the destination already exists it should be silently overwritten.
        """
        if dest is None:
            dest = src

        if src == dest and store == dest_store:
            raise ValueError("Source and destination cannot be the same asset in the same store.")

        raw_data = await self.get_raw_async(src, store)
        return await self.add_raw_async(dest.asset_type, dest.asset_id, raw_data, dest_store)

    def copy(
        self,
        src: AssetUri,
        dest_store: str,
        store: str = "local",
        dest: Optional[AssetUri] = None,
    ) -> AssetUri:
        """
        Copies an asset from one location to another synchronously.

        Performs a shallow copy by wrapping get_raw and add_raw.
        If dest is None, it defaults to the uri given in src.
        An assertion is raised if src and store are the same as dest and
        dest_store.
        If the destination already exists it should be silently overwritten.
        """
        return asyncio.run(self.copy_async(src, dest_store, store, dest))

    async def deepcopy_async(
        self,
        src: AssetUri,
        dest_store: str,
        store: str = "local",
        dest: Optional[AssetUri] = None,
    ) -> AssetUri:
        """
        Asynchronously deep copies an asset and its dependencies from a source store
        to a destination store.

        Args:
            src: The AssetUri of the source asset.
            dest_store: The name of the destination store.
            store: The name of the source store (defaults to "local").
            dest: Optional. The new AssetUri for the top-level asset in the
                  destination store. If None, the original URI is used.

        Returns:
            The AssetUri of the copied top-level asset in the destination store.

        Raises:
            ValueError: If the source or destination store is not registered.
            FileNotFoundError: If the source asset is not found.
            RuntimeError: If a cyclic dependency is detected.
        """
        logger.debug(
            f"DeepcopyAsync URI '{src}' from store '{store}' to '{dest_store}'"
            f" with dest '{dest}'"
        )
        if dest is None:
            dest = src

        if store not in self.stores:
            raise ValueError(f"Source store '{store}' not registered.")
        if dest_store not in self.stores:
            raise ValueError(f"Destination store '{dest_store}' not registered.")
        if store == dest_store and src == dest:
            raise ValueError(f"File '{src}' cannot be copied to itself.")

        # Fetch the source asset and its dependencies recursively
        # Use a new set for visited_uris for this deepcopy operation
        construction_data = await self._fetch_asset_construction_data_recursive_async(
            src, [store], set(), depth=None
        )
        if construction_data is None:
            raise FileNotFoundError(f"Source asset '{src}' not found in store '{store}'.")

        # Collect all assets (including dependencies) in a flat list,
        # ensuring dependencies are processed before the assets that depend on them.
        assets_to_copy: List[_AssetConstructionData] = []

        def collect_assets(data: _AssetConstructionData):
            if data.dependencies_data is not None:
                for dep_data in data.dependencies_data.values():
                    if dep_data:  # Only collect if dependency data was successfully fetched
                        collect_assets(dep_data)
            assets_to_copy.append(data)

        collect_assets(construction_data)

        # Process assets in the collected order (dependencies first)
        dest_store: AssetStore = self.stores[dest_store]
        copied_uris: Set[AssetUri] = set()
        for asset_data in assets_to_copy:
            # Prevent duplicate processing of the same asset
            asset_uri = dest if asset_data.uri == src else asset_data.uri
            if asset_uri in copied_uris:
                logger.debug(
                    f"Dependency '{asset_uri}' already added to '{dest_store}'," " skipping copy."
                )
                continue
            copied_uris.add(asset_uri)

            # Check if the dependency already exists in the destination store
            # Dependencies should be skipped if they exist, top-level should be overwritten.
            exists_in_dest = await dest_store.exists(asset_uri)
            if exists_in_dest and asset_uri != src:
                logger.debug(
                    f"Dependency '{asset_uri}' already exists in '{dest_store}'," " skipping copy."
                )
                continue

            # Put the asset (or dependency) into the destination store
            # Pass the dependency_uri_map to the store's put method.
            if exists_in_dest:
                # If it was not skipped above, this is the top-level asset. Update it.
                logger.debug(f"Updating asset '{asset_uri}' in '{dest_store}'")
                dest = await dest_store.update(
                    asset_uri,
                    asset_data.raw_data,
                )
            else:
                # If it doesn't exist, or if it's a dependency that doesn't exist, create it
                logger.debug(f"Creating asset '{asset_uri}' in '{dest_store}'")
                logger.debug(f"Raw data before writing: {asset_data.raw_data}")  # Added log
                await dest_store.create(
                    asset_uri.asset_type,
                    asset_uri.asset_id,
                    asset_data.raw_data,
                )

        logger.debug(f"DeepcopyAsync completed for '{src}' to '{dest}'")
        return dest

    def deepcopy(
        self,
        src: AssetUri,
        dest_store: str,
        store: str = "local",
        dest: Optional[AssetUri] = None,
    ) -> AssetUri:
        """
        Synchronously deep copies an asset and its dependencies from a source store
        to a destination store.

        Args:
            src: The AssetUri of the source asset.
            dest_store: The name of the destination store.
            store: The name of the source store (defaults to "local").
            dest: Optional. The new AssetUri for the top-level asset in the
                  destination store. If None, the original URI is used.

        Returns:
            The AssetUri of the copied top-level asset in the destination store.

        Raises:
            ValueError: If the source or destination store is not registered.
            FileNotFoundError: If the source asset is not found.
            RuntimeError: If a cyclic dependency is detected.
        """
        logger.debug(
            f"Deepcopy URI '{src}' from store '{store}' to '{dest_store}'" f" with dest '{dest}'"
        )
        return asyncio.run(self.deepcopy_async(src, dest_store, store, dest))

    def add_file(
        self,
        asset_type: str,
        path: pathlib.Path,
        store: str = "local",
        asset_id: Optional[str] = None,
    ) -> AssetUri:
        """
        Convenience wrapper around add_raw().
        If asset_id is None, the path.stem is used as the id.
        """
        return self.add_raw(asset_type, asset_id or path.stem, path.read_bytes(), store=store)

    def delete(self, uri: Union[AssetUri, str], store: str = "local") -> None:
        logger.debug(f"Delete URI '{uri}' from store '{store}'")
        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri

        async def _do_delete_async():
            selected_store = self.stores[store]
            await selected_store.delete(asset_uri_obj)
            if store in self._cacheable_stores:
                self.asset_cache.invalidate_for_uri(str(asset_uri_obj))

        asyncio.run(_do_delete_async())

    async def delete_async(self, uri: Union[AssetUri, str], store: str = "local") -> None:
        logger.debug(f"DeleteAsync URI '{uri}' from store '{store}'")
        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri
        selected_store = self.stores[store]
        await selected_store.delete(asset_uri_obj)
        if store in self._cacheable_stores:
            self.asset_cache.invalidate_for_uri(str(asset_uri_obj))

    async def is_empty_async(self, asset_type: Optional[str] = None, store: str = "local") -> bool:
        """Checks if the asset store has any assets of a given type (asynchronous)."""
        logger.debug(f"IsEmptyAsync: type='{asset_type}', store='{store}'")
        logger.debug(
            f"IsEmptyAsync: Looking up store '{store}'. Available stores: {list(self.stores.keys())}"
        )
        selected_store = self.stores.get(store)
        if not selected_store:
            raise ValueError(f"No store registered for name: {store}")
        return await selected_store.is_empty(asset_type)

    def is_empty(self, asset_type: Optional[str] = None, store: str = "local") -> bool:
        """Checks if the asset store has any assets of a given type (synchronous wrapper)."""
        logger.debug(
            f"IsEmpty: type='{asset_type}', store='{store}' from T:{threading.current_thread().name}"
        )
        try:
            return asyncio.run(self.is_empty_async(asset_type, store))
        except Exception as e:
            logger.error(
                f"IsEmpty: Error for type='{asset_type}', store='{store}': {e}",
                exc_info=False,
            )  # Changed exc_info to False
            raise

    async def list_versions_async(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
    ) -> List[AssetUri]:
        """Lists available versions for a given asset URI (asynchronous)."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(f"ListVersionsAsync: uri='{uri}', stores='{stores_list}'")
        asset_uri_obj = AssetUri(uri) if isinstance(uri, str) else uri

        # Note: list_versions_async currently only supports a single store.
        # If listing versions across multiple stores is needed, this needs to be updated.
        # For now, list from the first store.
        list_store = stores_list[0] if stores_list else "local"
        logger.debug(
            f"ListVersionsAsync: Looking up store '{list_store}'. Available stores: {list(self.stores.keys())}"
        )
        selected_store = self.stores.get(list_store)
        if not selected_store:
            raise ValueError(f"No store registered for name: {list_store}")
        return await selected_store.list_versions(asset_uri_obj)

    def list_versions(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = "local",
    ) -> List[AssetUri]:
        """Lists available versions for a given asset URI (synchronous wrapper)."""
        stores_list = [store] if isinstance(store, str) else store
        logger.debug(
            f"ListVersions: uri='{uri}', stores='{stores_list}' from T:{threading.current_thread().name}"
        )
        try:
            return asyncio.run(self.list_versions_async(uri, stores_list))
        except Exception as e:
            logger.error(
                f"ListVersions: Error for uri='{uri}', stores='{stores_list}': {e}",
                exc_info=False,
            )
            return []  # Return empty list on error to satisfy type hint

    def get_registered_asset_types(self) -> List[str]:
        """Returns a list of registered asset type names."""
        return list(self._asset_classes.keys())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils

TOOLTIP = (
    """Example Post, using Path.Commands instead of Path.toGCode strings for Path G-code output."""
)

SHOW_EDITOR = True


def fmt(num):
    fnum = ""
    fnum += "%.3f" % (num)
    return fnum


def ffmt(num):
    fnum = ""
    fnum += "%.1f" % (num)
    return fnum


class saveVals(object):
    """save command info for modal output"""

    def __init__(self, command):
        self.com = command.Name
        self.params = command.Parameters

    def retVals(self):
        return self.com, self.params


def lineout(command, oldvals, modal):
    line = ""
    if modal and (oldvals.com == command.Name):
        line += ""
    else:
        line += str(command.Name)
    if command.Name == "M6":
        line += "T" + str(int(command.Parameters["T"]))
    if command.Name == "M3":
        line += "S" + str(ffmt(command.Parameters["S"]))
    if command.Name == "M4":
        line += "S" + str(ffmt(command.Parameters["S"]))
    if "X" in command.Parameters:
        line += "X" + str(fmt(command.Parameters["X"]))
    if "Y" in command.Parameters:
        line += "Y" + str(fmt(command.Parameters["Y"]))
    if "Z" in command.Parameters:
        line += "Z" + str(fmt(command.Parameters["Z"]))
    if "I" in command.Parameters:
        line += "I" + str(fmt(command.Parameters["I"]))
    if "J" in command.Parameters:
        line += "J" + str(fmt(command.Parameters["J"]))
    if "F" in command.Parameters:
        line += "F" + str(ffmt(command.Parameters["F"]))
    return line


def export(obj, filename, argstring):
    modal = True
    gcode = ""
    safetyblock1 = "G90G40G49\n"
    gcode += safetyblock1

    units = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units")
    if units.GetInt("UserSchema") == 0:
        firstcommand = Path.Command("G21")  # metric mode
    else:
        firstcommand = Path.Command("G20")  # inch mode
    oldvals = saveVals(firstcommand)  # save first command for modal use
    fp = obj[0]
    gcode += firstcommand.Name

    if hasattr(fp, "Path"):
        for c in PathUtils.getPathWithPlacement(fp).Commands:
            gcode += lineout(c, oldvals, modal) + "\n"
            oldvals = saveVals(c)
        gcode += "M2\n"
        gfile = open(filename, "w")
        gfile.write(gcode)
        gfile.close()
    else:
        FreeCAD.Console.PrintError("Select a path object and try again\n")

    if SHOW_EDITOR:
        FreeCAD.Console.PrintMessage("Editor Activated\n")
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        dia.exec_()





import FreeCAD
import unittest
from .TechDrawTestUtilities import createPageWithSVGTemplate


class DrawViewAnnotationTest(unittest.TestCase):
    def setUp(self):
        """Creates a page"""
        FreeCAD.newDocument("TDAnno")
        FreeCAD.setActiveDocument("TDAnno")
        FreeCAD.ActiveDocument = FreeCAD.getDocument("TDAnno")
        self.page = createPageWithSVGTemplate()

    def tearDown(self):
        FreeCAD.closeDocument("TDAnno")

    def testMakeAnnotation(self):
        """Tests if an annotation can be added to page"""
        anno = FreeCAD.ActiveDocument.addObject(
            "TechDraw::DrawViewAnnotation", "TestAnno"
        )
        s = "Different Text"
        sl = list()
        sl.append(s)
        anno.Text = sl
        anno.TextStyle = "Bold"
        self.page.addView(anno)
        anno.X = 30.0
        anno.Y = 150.0
        FreeCAD.ActiveDocument.recompute()

        self.assertTrue("Up-to-date" in anno.State)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

"""
AssetManager tests.
"""

import unittest
import asyncio
from unittest.mock import Mock
import pathlib
import tempfile
from typing import Any, Mapping, List, Type, Optional, cast
from Path.Tool.assets import (
    AssetManager,
    FileStore,
    Asset,
    AssetUri,
    MemoryStore,
    AssetSerializer,
    DummyAssetSerializer,
)


# Mock Asset class for testing
class MockAsset(Asset):
    asset_type: str = "mock_asset"

    def __init__(self, data: Any = None, id: str = "mock_id"):
        self._data = data
        self._id = id

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: Type[AssetSerializer]) -> List[AssetUri]:
        # Mock implementation doesn't use data or format for dependencies
        return []

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
        serializer: Type[AssetSerializer],
    ) -> "MockAsset":
        # Create instance with provided id
        return cls(data, id)

    def to_bytes(self, serializer: Type[AssetSerializer]) -> bytes:
        return self._data

    def get_id(self) -> str:
        return self._id

    def get_data(self) -> bytes:
        """Returns the raw data stored in the mock asset."""
        return self._data


# Mock Asset class with dependencies for testing deepcopy
class MockAssetWithDeps(Asset):
    asset_type: str = "mock_asset_with_deps"

    def __init__(
        self,
        data: Any = None,
        id: str = "mock_id",
        dependencies: Optional[Mapping[AssetUri, Asset]] = None,
    ):
        self._data = data
        self._id = id
        self._dependencies = dependencies or {}

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: Type[AssetSerializer]) -> List[AssetUri]:
        # Assuming data is a simple JSON string like '{"deps": ["uri1", "uri2"]}'
        try:
            import json

            data_str = data.decode("utf-8")
            data_dict = json.loads(data_str)
            dep_uris_str = data_dict.get("deps", [])
            return [AssetUri(uri_str) for uri_str in dep_uris_str]
        except Exception:
            return []

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
        serializer: Type[AssetSerializer],
    ) -> "MockAssetWithDeps":
        # Create instance with provided id and resolved dependencies
        return cls(data, id, dependencies)

    def to_bytes(self, serializer: Type[AssetSerializer]) -> bytes:
        # Serialize data and dependency URIs into a simple format
        try:
            import json

            dep_uri_strs = [str(uri) for uri in self._dependencies.keys()]
            data_dict = {"data": self._data.decode("utf-8"), "deps": dep_uri_strs}
            return json.dumps(data_dict).encode("utf-8")
        except Exception:
            return self._data  # Fallback if serialization fails

    def get_id(self) -> str:
        return self._id

    def get_data(self) -> bytes:
        """Returns the raw data stored in the mock asset."""
        return self._data

    def get_dependencies(self) -> Mapping[AssetUri, Asset]:
        """Returns the resolved dependencies."""
        return self._dependencies


class TestPathToolAssetManager(unittest.TestCase):
    def test_register_store(self):
        manager = AssetManager()
        mock_store_local = Mock()
        mock_store_local.name = "local"
        mock_store_remote = Mock()
        mock_store_remote.name = "remote"

        manager.register_store(mock_store_local)
        self.assertEqual(manager.stores["local"], mock_store_local)

        manager.register_store(mock_store_remote)
        self.assertEqual(manager.stores["remote"], mock_store_remote)

        # Test overwriting
        mock_store_local_new = Mock()
        mock_store_local_new.name = "local"
        manager.register_store(mock_store_local_new)
        self.assertEqual(manager.stores["local"], mock_store_local_new)

    def test_register_asset(self):
        manager = AssetManager()
        # Register the actual MockAsset class
        manager.register_asset(MockAsset, DummyAssetSerializer)
        self.assertEqual(manager._asset_classes[MockAsset.asset_type], MockAsset)

        # Test registering a different actual Asset class
        class AnotherMockAsset(Asset):
            asset_type: str = "another_mock_asset"

            @classmethod
            def dependencies(cls, data: bytes) -> List[AssetUri]:
                return []

            @classmethod
            def from_bytes(
                cls,
                data: bytes,
                id: str,
                dependencies: Optional[Mapping[AssetUri, Asset]],
                serializer: Type[AssetSerializer],
            ) -> "AnotherMockAsset":
                return cls()

            def to_bytes(self, serializer: Type[AssetSerializer]) -> bytes:
                return b""

            def get_id(self) -> str:
                return "another_mock_id"

        manager.register_asset(AnotherMockAsset, DummyAssetSerializer)
        self.assertEqual(manager._asset_classes[AnotherMockAsset.asset_type], AnotherMockAsset)

        # Test overwriting
        manager.register_asset(
            MockAsset, DummyAssetSerializer
        )  # Registering again should overwrite
        self.assertEqual(manager._asset_classes[MockAsset.asset_type], MockAsset)

        # Test registering non-Asset class
        with self.assertRaises(TypeError):

            class NotAnAsset(Asset):  # Inherit from Asset
                pass

            manager.register_asset(NotAnAsset, DummyAssetSerializer)

    def test_get(self):
        # Setup AssetManager with a real LocalStore and the MockAsset class
        with tempfile.TemporaryDirectory() as tmpdir:
            base_dir = pathlib.Path(tmpdir)
            local_store = FileStore("local", base_dir)
            manager = AssetManager()
            manager.register_store(local_store)

            # Register the MockAsset class
            manager.register_asset(MockAsset, DummyAssetSerializer)

            # Create a test asset file via AssetManager
            test_data = b"test asset data"
            test_uri = manager.add_raw(
                asset_type=MockAsset.asset_type,
                asset_id="dummy_id_get",
                data=test_data,
                store="local",
            )

            # Call AssetManager.get
            retrieved_object = cast(MockAsset, manager.get(test_uri))

            # Assert the retrieved object is an instance of MockAsset
            self.assertIsInstance(retrieved_object, MockAsset)
            self.assertEqual(retrieved_object.get_data(), test_data)

            # Test error handling for non-existent URI
            non_existent_uri = AssetUri.build(MockAsset.asset_type, "non_existent", "1")
            with self.assertRaises(FileNotFoundError):
                manager.get(non_existent_uri)

            # Test error handling for no asset class registered
            non_registered_uri = AssetUri.build("non_existent_type", "dummy_id", "1")
            # Need to create a dummy file for the store to find
            dummy_data = b"dummy"
            manager.add_raw(
                asset_type="non_existent_type", asset_id="dummy_id", data=dummy_data, store="local"
            )

            with self.assertRaises(ValueError) as cm:
                manager.get(non_registered_uri)
            self.assertIn("No asset class registered for URI:", str(cm.exception))

    def test_delete(self):
        # Setup AssetManager with a real LocalStore
        with tempfile.TemporaryDirectory() as tmpdir:
            base_dir = pathlib.Path(tmpdir)
            local_store = FileStore("local", base_dir)
            manager = AssetManager()
            manager.register_store(local_store)

            # Create a test asset file
            test_data = b"test asset data to delete"
            test_uri = manager.add_raw(
                asset_type="temp_asset", asset_id="dummy_id_delete", data=test_data, store="local"
            )
            test_path = base_dir / "temp_asset" / str(test_uri.asset_id) / str(test_uri.version)
            self.assertTrue(test_path.exists())

            # Call AssetManager.delete
            manager.delete(test_uri)

            # Verify file deletion
            self.assertFalse(test_path.exists())

            # Test error handling for non-existent URI (should not raise error
            # as LocalStore.delete handles this)
            non_existent_uri = AssetUri.build(
                "temp_asset", "non_existent", "1"  # Keep original for logging
            )
            manager.delete(non_existent_uri)  # Should not raise

    def test_create(self):
        # Setup AssetManager with LocalStore and MockAsset class
        with tempfile.TemporaryDirectory() as tmpdir:
            base_dir = pathlib.Path(tmpdir)
            local_store = FileStore("local", base_dir)
            manager = AssetManager()
            manager.register_store(local_store)

            # Register the MockAsset class
            manager.register_asset(MockAsset, DummyAssetSerializer)

            # Create a MockAsset instance with a specific id
            test_obj = MockAsset(b"object data", id="mocked_asset_id")

            # Call manager.add to create
            created_uri = manager.add(test_obj, store="local")

            # Assert returned URI is as expected
            expected_uri = AssetUri.build(MockAsset.asset_type, "mocked_asset_id", "1")
            self.assertEqual(created_uri, expected_uri)

            # Verify the asset was created
            retrieved_data = asyncio.run(local_store.get(created_uri))
            self.assertEqual(retrieved_data, test_obj.get_data())

            # Test error handling (store not found)
            with self.assertRaises(ValueError) as cm:
                manager.add(test_obj, store="non_existent_store")
            self.assertIn("No store registered for name:", str(cm.exception))

        with tempfile.TemporaryDirectory() as tmpdir:
            local_store = MemoryStore("local")
            manager = AssetManager()
            manager.register_store(local_store)

            # Register the MockAsset class
            manager.register_asset(MockAsset, DummyAssetSerializer)

            # First, create an asset
            initial_data = b"initial data"
            asset_id = "some_asset_id"
            test_uri = manager.add_raw(MockAsset.asset_type, asset_id, initial_data, "local")
            self.assertEqual(test_uri.version, "1")

            # Create a MockAsset instance with the same id for update
            updated_data = b"updated object data"
            test_obj = MockAsset(updated_data, id=asset_id)

            # Call manager.add to update
            updated_uri = manager.add(test_obj, store="local")

            # Assert returned URI matches the original except for version
            self.assertEqual(updated_uri.asset_type, test_uri.asset_type)
            self.assertEqual(updated_uri.asset_id, test_uri.asset_id)
            self.assertEqual(updated_uri.version, "2")

            # Verify the asset was updated
            obj = cast(MockAsset, manager.get(updated_uri, store="local"))
            self.assertEqual(updated_data, test_obj.get_data())
            self.assertIsInstance(obj, MockAsset)
            self.assertEqual(updated_data, obj.get_data())

            # Test error handling (store not found)
            with self.assertRaises(ValueError) as cm:
                manager.add(test_obj, store="non_existent_store")
            self.assertIn("No store registered for name:", str(cm.exception))

    def test_create_raw(self):
        # Setup AssetManager with a real MemoryStore
        memory_store = MemoryStore("memory_raw")
        manager = AssetManager()
        manager.register_store(memory_store)

        asset_type = "raw_test_type"
        asset_id = "raw_test_id"
        data = b"raw test data"

        # Expected URI with version 1 (assuming MemoryStore uses integer versions)
        expected_uri = AssetUri.build(asset_type, asset_id, "1")

        # Call manager.add_raw
        created_uri = manager.add_raw(
            asset_type=asset_type, asset_id=asset_id, data=data, store="memory_raw"
        )

        # Assert returned URI is correct (check asset_type and asset_id)
        self.assertEqual(created_uri.asset_type, asset_type)
        self.assertEqual(created_uri.asset_id, asset_id)
        self.assertEqual(created_uri, expected_uri)

        # Verify data was stored using the actual created_uri
        # Await the async get method using asyncio.run
        retrieved_data = asyncio.run(memory_store.get(created_uri))
        self.assertEqual(retrieved_data, data)

        # Test error handling (store not found)
        with self.assertRaises(ValueError) as cm:
            manager.add_raw(
                asset_type=asset_type, asset_id=asset_id, data=data, store="non_existent_store"
            )
        self.assertIn("No store registered for name:", str(cm.exception))

    def test_get_raw(self):
        # Setup AssetManager with a real MemoryStore
        memory_store = MemoryStore("memory_raw_get")
        manager = AssetManager()
        manager.register_store(memory_store)

        test_uri_str = "test_type://test_id/1"
        test_uri = AssetUri(test_uri_str)
        expected_data = b"retrieved raw data"

        # Manually put data into the memory store
        manager.add_raw("test_type", "test_id", expected_data, "memory_raw_get")

        # Call manager.get_raw using the URI returned by add_raw
        retrieved_data = manager.get_raw(test_uri, store="memory_raw_get")

        # Assert returned data matches store's result
        self.assertEqual(retrieved_data, expected_data)

        # Test error handling (store not found)
        non_existent_uri = AssetUri("type://id/1")
        # Test error handling (asset not found in any store, including non-existent ones)
        non_existent_uri = AssetUri("type://id/1")
        with self.assertRaises(FileNotFoundError) as cm:
            manager.get_raw(non_existent_uri, store="non_existent_store")
        self.assertIn(
            "Asset 'type://id/1' not found in stores '['non_existent_store']'", str(cm.exception)
        )

    def test_is_empty(self):
        # Setup AssetManager with a real MemoryStore
        memory_store = MemoryStore("memory_empty")
        manager = AssetManager()
        manager.register_store(memory_store)

        # Test when store is empty
        self.assertTrue(manager.is_empty(store="memory_empty"))

        # Add an asset and test again
        manager.add_raw("test_type", "test_id", b"data", "memory_empty")
        self.assertFalse(manager.is_empty(store="memory_empty"))

        # Test with asset type
        self.assertTrue(manager.is_empty(store="memory_empty", asset_type="another_type"))
        self.assertFalse(manager.is_empty(store="memory_empty", asset_type="test_type"))

        # Test error handling (store not found)
        with self.assertRaises(ValueError) as cm:
            manager.is_empty(store="non_existent_store")
        self.assertIn("No store registered for name:", str(cm.exception))

    def test_count_assets(self):
        # Setup AssetManager with a real MemoryStore
        memory_store = MemoryStore("memory_count")
        manager = AssetManager()
        manager.register_store(memory_store)

        # Test when store is empty
        self.assertEqual(manager.count_assets(store="memory_count"), 0)
        self.assertEqual(manager.count_assets(store="memory_count", asset_type="type1"), 0)

        # Add assets and test counts
        manager.add_raw("type1", "asset1", b"data1", "memory_count")
        self.assertEqual(manager.count_assets(store="memory_count"), 1)
        self.assertEqual(manager.count_assets(store="memory_count", asset_type="type1"), 1)
        self.assertEqual(manager.count_assets(store="memory_count", asset_type="type2"), 0)

        manager.add_raw("type2", "asset2", b"data2", "memory_count")
        manager.add_raw("type1", "asset3", b"data3", "memory_count")
        self.assertEqual(manager.count_assets(store="memory_count"), 3)
        self.assertEqual(manager.count_assets(store="memory_count", asset_type="type1"), 2)
        self.assertEqual(manager.count_assets(store="memory_count", asset_type="type2"), 1)

        # Test error handling (store not found)
        with self.assertRaises(ValueError) as cm:
            manager.count_assets(store="non_existent_store")
        self.assertIn("No store registered for name:", str(cm.exception))

    def test_get_bulk(self):
        # Setup AssetManager with a real MemoryStore and MockAsset class
        memory_store = MemoryStore("memory_bulk")
        manager = AssetManager()
        manager.register_store(memory_store)
        manager.register_asset(MockAsset, DummyAssetSerializer)

        # Create some assets in the memory store
        data1 = b"data for id1"
        data2 = b"data for id2"
        uri1 = manager.add_raw(MockAsset.asset_type, "id1", data1, "memory_bulk")
        uri2 = manager.add_raw(MockAsset.asset_type, "id2", data2, "memory_bulk")
        uri3 = AssetUri.build(MockAsset.asset_type, "non_existent", "1")
        uris = [uri1, uri2, uri3]

        # Call manager.get_bulk
        retrieved_assets = cast(List[MockAsset], manager.get_bulk(uris, store="memory_bulk"))

        # Assert the correct number of assets were returned
        self.assertEqual(len(retrieved_assets), 3)

        # Assert the retrieved assets are MockAsset instances with correct data
        self.assertIsInstance(retrieved_assets[0], MockAsset)
        self.assertEqual(retrieved_assets[0].get_data(), data1)

        self.assertIsInstance(retrieved_assets[1], MockAsset)
        self.assertEqual(retrieved_assets[1].get_data(), data2)

        # Assert the non-existent asset is None
        self.assertIsNone(retrieved_assets[2])

        # Test error handling (store not found)
        # Test handling of non-existent store (should skip and not raise ValueError)
        # The test already asserts that the non-existent asset is None, which is the expected behavior.
        manager.get_bulk(uris, store="non_existent_store")

    def test_fetch(self):
        # Setup AssetManager with a real MemoryStore and MockAsset class
        memory_store = MemoryStore("memory_fetch")
        manager = AssetManager()
        manager.register_store(memory_store)
        manager.register_asset(MockAsset, DummyAssetSerializer)

        # Create some assets in the memory store
        data1 = b"data for id1"
        data2 = b"data for id2"
        manager.add_raw(MockAsset.asset_type, "id1", data1, "memory_fetch")
        manager.add_raw(MockAsset.asset_type, "id2", data2, "memory_fetch")
        # Create an asset of a different type
        manager.add_raw("another_type", "id3", b"data for id3", "memory_fetch")
        AssetUri.build(MockAsset.asset_type, "non_existent", "1")

        # Call manager.fetch without filters
        # This should raise ValueError because uri3 has an unregistered type
        with self.assertRaises(ValueError) as cm:
            manager.fetch(store="memory_fetch")
        self.assertIn("No asset class registered for URI:", str(cm.exception))

        # Now test fetching with a registered asset type filter
        # Setup a new manager and store to avoid state from previous test
        memory_store_filtered = MemoryStore("memory_fetch_filtered")
        manager_filtered = AssetManager()
        manager_filtered.register_store(memory_store_filtered)
        manager_filtered.register_asset(MockAsset, DummyAssetSerializer)

        # Create assets again
        manager_filtered.add_raw(MockAsset.asset_type, "id1", data1, "memory_fetch_filtered")
        manager_filtered.add_raw(MockAsset.asset_type, "id2", data2, "memory_fetch_filtered")
        manager_filtered.add_raw("another_type", "id3", b"data for id3", "memory_fetch_filtered")

        retrieved_assets_filtered = cast(
            List[MockAsset],
            manager_filtered.fetch(asset_type=MockAsset.asset_type, store="memory_fetch_filtered"),
        )

        # Assert the correct number of assets were returned
        self.assertEqual(len(retrieved_assets_filtered), 2)

        # Assert the retrieved assets are MockAsset instances with correct data
        self.assertIsInstance(retrieved_assets_filtered[0], MockAsset)
        self.assertEqual(
            retrieved_assets_filtered[0].get_data().decode("utf-8"),
            data1.decode("utf-8"),
        )

        self.assertIsInstance(retrieved_assets_filtered[1], MockAsset)
        self.assertEqual(
            retrieved_assets_filtered[1].get_data().decode("utf-8"),
            data2.decode("utf-8"),
        )

        # Test error handling (store not found)
        with self.assertRaises(ValueError) as cm:
            manager.fetch(store="non_existent_store")
        self.assertIn("No store registered for name:", str(cm.exception))

    def test_copy(self):
        # Setup AssetManager with two MemoryStores and MockAsset class
        memory_store_src = MemoryStore("memory_copy_src")
        memory_store_dest = MemoryStore("memory_copy_dest")
        manager = AssetManager()
        manager.register_store(memory_store_src)
        manager.register_store(memory_store_dest)
        manager.register_asset(MockAsset, DummyAssetSerializer)

        # Create a source asset
        src_data = b"source asset data"
        src_uri = manager.add_raw(MockAsset.asset_type, "source_id", src_data, "memory_copy_src")

        # Test copying to a different store with default destination URI
        copied_uri_default_dest = manager.copy(
            src_uri, dest_store="memory_copy_dest", store="memory_copy_src"
        )
        self.assertEqual(copied_uri_default_dest.asset_type, src_uri.asset_type)
        self.assertEqual(copied_uri_default_dest.asset_id, src_uri.asset_id)
        self.assertEqual(copied_uri_default_dest.version, "1")  # First version in dest

        # Verify the copied asset exists in the destination store
        copied_data_default_dest = manager.get_raw(
            copied_uri_default_dest, store="memory_copy_dest"
        )
        self.assertEqual(copied_data_default_dest, src_data)

        # Test copying to a different store with a specified destination URI
        dest_uri_specified = AssetUri.build(MockAsset.asset_type, "specified_dest_id", "1")
        copied_uri_specified_dest = manager.copy(
            src_uri,
            dest_store="memory_copy_dest",
            store="memory_copy_src",
            dest=dest_uri_specified,
        )
        self.assertEqual(copied_uri_specified_dest, dest_uri_specified)

        # Verify the copied asset exists at the specified destination URI
        copied_data_specified_dest = manager.get_raw(
            copied_uri_specified_dest, store="memory_copy_dest"
        )
        self.assertEqual(copied_data_specified_dest, src_data)

        # Test copying to the same store with a different destination URI
        dest_uri_same_store = AssetUri.build(MockAsset.asset_type, "same_store_dest", "1")
        copied_uri_same_store = manager.copy(
            src_uri, dest_store="memory_copy_src", store="memory_copy_src", dest=dest_uri_same_store
        )
        self.assertEqual(copied_uri_same_store, dest_uri_same_store)

        # Verify the copied asset exists in the same store at the new URI
        copied_data_same_store = manager.get_raw(copied_uri_same_store, store="memory_copy_src")
        self.assertEqual(copied_data_same_store, src_data)

        # Test assertion for source and destination being the same
        with self.assertRaises(ValueError) as cm:
            manager.copy(
                src_uri, dest_store="memory_copy_src", store="memory_copy_src", dest=src_uri
            )
        self.assertIn(
            "Source and destination cannot be the same asset in the same store.",
            str(cm.exception),
        )

        # Test overwriting existing destination (add a different asset at specified_dest_id)
        overwrite_data = b"data to be overwritten"
        overwrite_uri = manager.add_raw(
            MockAsset.asset_type, "specified_dest_id", overwrite_data, "memory_copy_dest"
        )
        self.assertEqual(overwrite_uri.version, "2")  # Should be version 2 now

        # Copy the original src_uri to the existing destination
        copied_uri_overwrite = manager.copy(
            src_uri,
            dest_store="memory_copy_dest",
            store="memory_copy_src",
            dest=dest_uri_specified,
        )
        # The version should be incremented again due to overwrite
        self.assertEqual(copied_uri_overwrite.version, "3")

        # Verify the destination now contains the source data
        retrieved_overwritten_data = manager.get_raw(copied_uri_overwrite, store="memory_copy_dest")
        self.assertEqual(retrieved_overwritten_data, src_data)

    def test_deepcopy(self):
        # Setup AssetManager with two MemoryStores and MockAssetWithDeps class
        memory_store_src = MemoryStore("memory_deepcopy_src")
        memory_store_dest = MemoryStore("memory_deepcopy_dest")
        manager = AssetManager()
        manager.register_store(memory_store_src)
        manager.register_store(memory_store_dest)
        manager.register_asset(MockAssetWithDeps, DummyAssetSerializer)

        # Create dependency assets in the source store
        dep1_data = b'{"data": "dependency 1 data", "deps": []}'
        dep2_data = b'{"data": "dependency 2 data", "deps": []}'
        dep1_uri = manager.add_raw(
            MockAssetWithDeps.asset_type, "dep1_id", dep1_data, "memory_deepcopy_src"
        )
        dep2_uri = manager.add_raw(
            MockAssetWithDeps.asset_type, "dep2_id", dep2_data, "memory_deepcopy_src"
        )

        # Create a source asset with dependencies
        src_data = (
            b'{"data": "source asset data", "deps": ["'
            + str(dep1_uri).encode("utf-8")
            + b'", "'
            + str(dep2_uri).encode("utf-8")
            + b'"]}'
        )
        src_uri = manager.add_raw(
            MockAssetWithDeps.asset_type, "source_id", src_data, "memory_deepcopy_src"
        )

        # Test deep copying to a different store with default destination URI
        copied_uri_default_dest = manager.deepcopy(
            src_uri, dest_store="memory_deepcopy_dest", store="memory_deepcopy_src"
        )
        self.assertEqual(copied_uri_default_dest.asset_type, src_uri.asset_type)
        self.assertEqual(copied_uri_default_dest.asset_id, src_uri.asset_id)
        self.assertEqual(copied_uri_default_dest.version, "1")  # First version in dest

        # Verify the copied top-level asset exists in the destination store
        copied_asset_default_dest = cast(
            MockAssetWithDeps, manager.get(copied_uri_default_dest, store="memory_deepcopy_dest")
        )
        self.assertIsInstance(copied_asset_default_dest, MockAssetWithDeps)
        # The copied asset's data should be the serialized form including dependencies
        expected_data = b'{"data": "source asset data", "deps": ["mock_asset_with_deps://dep1_id/1", "mock_asset_with_deps://dep2_id/1"]}'
        self.assertEqual(copied_asset_default_dest.get_data(), expected_data)

        # Verify dependencies were also copied and resolved correctly
        copied_deps_default_dest = copied_asset_default_dest.get_dependencies()
        self.assertEqual(len(copied_deps_default_dest), 2)
        self.assertIn(dep1_uri, copied_deps_default_dest)
        self.assertIn(dep2_uri, copied_deps_default_dest)

        copied_dep1 = cast(MockAssetWithDeps, copied_deps_default_dest[dep1_uri])
        self.assertIsInstance(copied_dep1, MockAssetWithDeps)
        self.assertEqual(copied_dep1.get_data(), b'{"data": "dependency 1 data", "deps": []}')

        copied_dep2 = cast(MockAssetWithDeps, copied_deps_default_dest[dep2_uri])
        self.assertIsInstance(copied_dep2, MockAssetWithDeps)
        self.assertEqual(copied_dep2.get_data(), b'{"data": "dependency 2 data", "deps": []}')

        # Test deep copying with a specified destination URI for the top-level asset
        dest_uri_specified = AssetUri.build(MockAssetWithDeps.asset_type, "specified_dest_id", "1")
        copied_uri_specified_dest = manager.deepcopy(
            src_uri,
            dest_store="memory_deepcopy_dest",
            store="memory_deepcopy_src",
            dest=dest_uri_specified,
        )
        self.assertEqual(copied_uri_specified_dest, dest_uri_specified)

        # Verify the copied asset exists at the specified destination URI
        copied_asset_specified_dest = cast(
            MockAssetWithDeps, manager.get(copied_uri_specified_dest, store="memory_deepcopy_dest")
        )
        self.assertIsInstance(copied_asset_specified_dest, MockAssetWithDeps)
        self.assertEqual(
            copied_asset_specified_dest.get_data(),
            b'{"data": "source asset data", "deps": ["mock_asset_with_deps://dep1_id/1", "mock_asset_with_deps://dep2_id/1"]}',
        )

        # Verify dependencies were copied and resolved correctly (their URIs should be
        # in the destination store, but their asset_type and asset_id should be the same)
        copied_deps_specified_dest = copied_asset_specified_dest.get_dependencies()
        self.assertEqual(len(copied_deps_specified_dest), 2)

        # The keys in the dependencies mapping should be the *original* URIs,
        # but the values should be the *copied* dependency assets.
        self.assertIn(dep1_uri, copied_deps_specified_dest)
        self.assertIn(dep2_uri, copied_deps_specified_dest)

        copied_dep1_specified = cast(MockAssetWithDeps, copied_deps_specified_dest[dep1_uri])
        self.assertIsInstance(copied_dep1_specified, MockAssetWithDeps)
        self.assertEqual(
            copied_dep1_specified.get_data(), b'{"data": "dependency 1 data", "deps": []}'
        )
        # Check the URI of the copied dependency in the destination store
        self.assertIsNotNone(
            manager.get_or_none(copied_dep1_specified.get_uri(), store="memory_deepcopy_dest")
        )
        self.assertEqual(copied_dep1_specified.get_uri().asset_type, dep1_uri.asset_type)
        self.assertEqual(copied_dep1_specified.get_uri().asset_id, dep1_uri.asset_id)

        copied_dep2_specified = cast(MockAssetWithDeps, copied_deps_specified_dest[dep2_uri])
        self.assertIsInstance(copied_dep2_specified, MockAssetWithDeps)
        self.assertEqual(
            copied_dep2_specified.get_data(), b'{"data": "dependency 2 data", "deps": []}'
        )
        # Check the URI of the copied dependency in the destination store
        self.assertIsNotNone(
            manager.get_or_none(copied_dep2_specified.get_uri(), store="memory_deepcopy_dest")
        )
        self.assertEqual(copied_dep2_specified.get_uri().asset_type, dep2_uri.asset_type)
        self.assertEqual(copied_dep2_specified.get_uri().asset_id, dep2_uri.asset_id)

        # Test handling of existing dependencies in the destination store (should be skipped)
        # Add a dependency with the same URI as dep1_uri to the destination store
        existing_dep1_data = b'{"data": "existing dependency 1 data", "deps": []}'
        existing_dep1_uri_in_dest = manager.add_raw(
            dep1_uri.asset_type, dep1_uri.asset_id, existing_dep1_data, "memory_deepcopy_dest"
        )
        self.assertEqual(existing_dep1_uri_in_dest.version, "2")  # Should be version 2 now

        # Deep copy the source asset again
        copied_uri_existing_dep = manager.deepcopy(
            src_uri, dest_store="memory_deepcopy_dest", store="memory_deepcopy_src"
        )
        # The top-level asset should be overwritten, incrementing its version
        self.assertEqual(copied_uri_existing_dep.version, "2")

        # Verify the top-level asset was overwritten
        copied_asset_existing_dep = cast(
            MockAssetWithDeps, manager.get(copied_uri_existing_dep, store="memory_deepcopy_dest")
        )
        self.assertIsInstance(copied_asset_existing_dep, MockAssetWithDeps)
        self.assertEqual(
            copied_asset_existing_dep.get_data(),
            b'{"data": "source asset data", "deps": ["mock_asset_with_deps://dep1_id/1", "mock_asset_with_deps://dep2_id/1"]}',
        )

        # Verify that the existing dependency was *not* overwritten
        retrieved_existing_dep1 = manager.get_raw(
            existing_dep1_uri_in_dest, store="memory_deepcopy_dest"
        )
        self.assertEqual(retrieved_existing_dep1, existing_dep1_data)

        # Verify the dependencies in the copied asset still point to the correct
        # (existing) dependency in the destination store.
        copied_deps_existing_dep = copied_asset_existing_dep.get_dependencies()
        self.assertEqual(len(copied_deps_existing_dep), 2)
        self.assertIn(dep1_uri, copied_deps_existing_dep)
        self.assertIn(dep2_uri, copied_deps_existing_dep)

        copied_dep1_existing = cast(MockAssetWithDeps, copied_deps_existing_dep[dep1_uri])
        self.assertIsInstance(copied_dep1_existing, MockAssetWithDeps)
        self.assertEqual(
            copied_dep1_existing.get_data(), b'{"data": "dependency 1 data", "deps": []}'
        )  # Should be the original data from source

        copied_dep2_existing = cast(MockAssetWithDeps, copied_deps_existing_dep[dep2_uri])
        self.assertIsInstance(copied_dep2_existing, MockAssetWithDeps)
        self.assertEqual(
            copied_dep2_existing.get_data(), b'{"data": "dependency 2 data", "deps": []}'
        )  # Should be the newly copied raw data

        # Test handling of existing top-level asset in the destination store (should be overwritten)
        # This was implicitly tested in the previous step where the top-level asset's
        # version was incremented. Let's add a more explicit test.
        overwrite_src_data = b'{"data": "overwrite source data", "deps": []}'
        overwrite_src_uri = manager.add_raw(
            MockAssetWithDeps.asset_type,
            "overwrite_source_id",
            overwrite_src_data,
            "memory_deepcopy_src",
        )

        # Add an asset to the destination store with the same URI as overwrite_src_uri
        existing_dest_data = b'{"data": "existing destination data", "deps": []}'
        existing_dest_uri = manager.add_raw(
            overwrite_src_uri.asset_type,
            overwrite_src_uri.asset_id,
            existing_dest_data,
            "memory_deepcopy_dest",
        )
        self.assertEqual(existing_dest_uri.version, "1")

        # Deep copy overwrite_src_uri to the existing destination URI
        copied_uri_overwrite_top = manager.deepcopy(
            overwrite_src_uri,
            dest_store="memory_deepcopy_dest",
            store="memory_deepcopy_src",
            dest=existing_dest_uri,
        )
        # The version should be incremented
        self.assertEqual(copied_uri_overwrite_top.version, "2")

        # Verify the destination now contains the source data
        retrieved_overwritten_top = manager.get_raw(
            copied_uri_overwrite_top, store="memory_deepcopy_dest"
        )
        # Need to parse the data to get the actual content
        import json

        retrieved_data_dict = json.loads(retrieved_overwritten_top.decode("utf-8"))
        self.assertEqual(retrieved_data_dict.get("data"), b"overwrite source data".decode("utf-8"))

        # Test error handling for non-existent source asset
        non_existent_src_uri = AssetUri.build(MockAssetWithDeps.asset_type, "non_existent_src", "1")
        with self.assertRaises(FileNotFoundError) as cm:
            manager.deepcopy(
                non_existent_src_uri, dest_store="memory_deepcopy_dest", store="memory_deepcopy_src"
            )
        self.assertIn("Source asset", str(cm.exception))
        self.assertIn("not found", str(cm.exception))

        # Test error handling for non-existent source store
        with self.assertRaises(ValueError) as cm:
            manager.deepcopy(src_uri, dest_store="memory_deepcopy_dest", store="non_existent_store")
        self.assertIn("Source store", str(cm.exception))
        self.assertIn("not registered", str(cm.exception))

        # Test error handling for non-existent destination store
        with self.assertRaises(ValueError) as cm:
            manager.deepcopy(src_uri, dest_store="non_existent_store", store="memory_deepcopy_src")
        self.assertIn("Destination store", str(cm.exception))
        self.assertIn("not registered", str(cm.exception))

    def test_exists(self):
        # Setup AssetManager with a MemoryStore
        memory_store = MemoryStore("memory_exists")
        manager = AssetManager()
        manager.register_store(memory_store)

        # Create an asset
        test_uri = manager.add_raw("test_type", "test_id", b"data", "memory_exists")

        # Test exists for an existing asset
        self.assertTrue(manager.exists(test_uri, store="memory_exists"))

        # Test exists for a non-existent asset
        non_existent_uri = AssetUri.build("test_type", "non_existent_id", "1")
        self.assertFalse(manager.exists(non_existent_uri, store="memory_exists"))

        # Test exists for a non-existent store (should raise ValueError)
        with self.assertRaises(ValueError) as cm:
            manager.exists(test_uri, store="non_existent_store")
        self.assertIn("No store registered for name:", str(cm.exception))


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Facebinder objects.

A facebinder is a surface or shell created from the face of a solid object.
This tool allows extracting such faces to be used for other purposes
including extruding solids from faces.
"""
## @package gui_facebinders
# \ingroup draftguitools
# \brief Provides GUI tools to create Facebinder objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils

from draftutils.messages import _msg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Facebinder(gui_base_original.Creator):
    """Gui Command for the Facebinder tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Facebinder",
            "Accel": "F,F",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Facebinder", "Facebinder"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Facebinder", "Creates a facebinder from the selected faces"
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Facebinder")
        if not self.ui:
            return
        if not Gui.Selection.getSelection():
            self.ui.selectUi(on_close_call=self.finish)
            _msg(translate("draft", "Select faces from existing objects"))
            self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed when a valid selection has been made."""
        if self.call is not None:
            self.end_callbacks(self.call)
        if Gui.Selection.getSelection():
            App.ActiveDocument.openTransaction("Create Facebinder")
            Gui.addModule("Draft")
            Gui.doCommand("sels = FreeCADGui.Selection.getSelectionEx('', 0)")
            Gui.doCommand("facebinder = Draft.make_facebinder(sels)")
            Gui.doCommand("Draft.autogroup(facebinder)")
            Gui.doCommand("FreeCAD.ActiveDocument.recompute()")
            App.ActiveDocument.commitTransaction()
            App.ActiveDocument.recompute()
        self.finish()


Draft_Facebinder = Facebinder
Gui.addCommand("Draft_Facebinder", Facebinder())

## @}
